import collections
import logging
import traceback
from functools import partial
from typing import List, Dict, Optional

from PySide2 import QtCore, QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QRadioButton, QGroupBox, QHBoxLayout, QLabel, QPushButton, QVBoxLayout, \
    QApplication, QDialog, QAction, QMenu
from asyncqt import asyncSlot

from randovania.game_description import data_reader, default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import PickupNode
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.games.prime import patcher_file
from randovania.gui.dialog.echoes_user_preferences_dialog import EchoesUserPreferencesDialog
from randovania.gui.dialog.game_input_dialog import GameInputDialog
from randovania.gui.generated.seed_details_window_ui import Ui_SeedDetailsWindow
from randovania.gui.lib import preset_describer, async_dialog, common_qt_lib
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.gui.lib.close_event_widget import CloseEventWidget
from randovania.gui.lib.common_qt_lib import set_default_window_icon, prompt_user_for_output_game_log
from randovania.gui.lib.window_manager import WindowManager
from randovania.interface_common import simplified_patcher, status_update_lib
from randovania.interface_common.options import Options, InfoAlert
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.interface_common.status_update_lib import ProgressUpdateCallable
from randovania.layout.layout_description import LayoutDescription


def _unique(iterable):
    seen = set()

    for item in iterable:
        if item in seen:
            continue
        seen.add(item)
        yield item


def _show_pickup_spoiler(button):
    target_player = getattr(button, "target_player", None)
    if target_player is not None:
        label = f"{button.item_name} for {button.player_names[target_player]}"
    else:
        label = button.item_name
    button.setText(label)
    button.item_is_hidden = False


def _hide_pickup_spoiler(button):
    button.setText("Hidden")
    button.item_is_hidden = True


class SeedDetailsWindow(CloseEventWidget, Ui_SeedDetailsWindow, BackgroundTaskMixin):
    _on_bulk_change: bool = False
    _history_items: List[QRadioButton]
    pickup_spoiler_buttons: List[QPushButton]
    layout_description: LayoutDescription
    _options: Options
    _window_manager: Optional[WindowManager]
    _player_names: Dict[int, str]
    _pickup_spoiler_current_game: Optional[GameDescription] = None
    _last_percentage: float = 0

    def __init__(self, window_manager: Optional[WindowManager], options: Options):
        super().__init__()
        self.setupUi(self)
        set_default_window_icon(self)

        self._history_items = []
        self.pickup_spoiler_buttons = []
        self._pickup_spoiler_world_to_group = {}
        self._options = options
        self._window_manager = window_manager

        # Ui
        self._tool_button_menu = QMenu(self.tool_button)
        self.tool_button.setMenu(self._tool_button_menu)

        self._action_open_tracker = QAction(self)
        self._action_open_tracker.setText("Open map tracker")
        self._action_open_tracker.setEnabled(self._window_manager is not None)
        self._tool_button_menu.addAction(self._action_open_tracker)

        self._action_copy_permalink = QAction(self)
        self._action_copy_permalink.setText("Copy Permalink")
        self._tool_button_menu.addAction(self._action_copy_permalink)

        self._action_open_dolphin = QAction(self)
        self._action_open_dolphin.setText("Open Dolphin Hook")
        self._tool_button_menu.addAction(self._action_open_dolphin)
        self._action_open_dolphin.setVisible(False)

        # Signals
        self.export_log_button.clicked.connect(self._export_log)
        self.export_iso_button.clicked.connect(self._export_iso)
        self._action_open_tracker.triggered.connect(self._open_map_tracker)
        self._action_copy_permalink.triggered.connect(self._copy_permalink)
        self.pickup_spoiler_pickup_combobox.currentTextChanged.connect(self._on_change_pickup_filter)
        self.pickup_spoiler_show_all_button.clicked.connect(self._toggle_show_all_pickup_spoiler)
        self.player_index_combo.activated.connect(self._update_current_player)
        self.CloseEvent.connect(self.stop_background_process)

        # Progress
        self.background_tasks_button_lock_signal.connect(self.enable_buttons_with_background_tasks)
        self.progress_update_signal.connect(self.update_progress)
        self.stop_background_process_button.clicked.connect(self.stop_background_process)

        # Cosmetic
        self.customize_user_preferences_button.clicked.connect(self._open_user_preferences_dialog)

    @property
    def current_player_index(self) -> int:
        return self.player_index_combo.currentData()

    # Operations
    def _copy_permalink(self):
        QApplication.clipboard().setText(self.layout_description.permalink.as_base64_str)

    def _export_log(self):
        game = self.layout_description.permalink.get_preset(self.current_player_index).configuration.game
        default_name = "{} Randomizer - {}.{}".format(game.short_name,
                                                      self.layout_description.shareable_word_hash,
                                                      self.layout_description.file_extension())
        json_path = prompt_user_for_output_game_log(self, default_name=default_name)
        if json_path is not None:
            self.layout_description.save_to_file(json_path)

    async def _show_dialog_for_prime3_layout(self):
        from randovania.games.prime import gollop_corruption_patcher

        patches = self.layout_description.all_patches[self.current_player_index]
        game = default_database.game_description_for(RandovaniaGame.PRIME3)

        item_names = []
        for index in range(game.world_list.num_pickup_nodes):
            p_index = PickupIndex(index)
            if p_index in patches.pickup_assignment:
                name = patches.pickup_assignment[p_index].pickup.name
            else:
                name = "Missile Expansion"
            item_names.append(name)

        layout_string = gollop_corruption_patcher.layout_string_for_items(item_names)
        starting_location = patches.starting_location

        commands = "\n".join([
            f'set seed="{layout_string}"',
            f'set "starting_items={gollop_corruption_patcher.starting_items_for(patches.starting_items)}"',
            f'set "starting_location={gollop_corruption_patcher.starting_location_for(starting_location)}"',
        ])
        message_box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information, "Commands for patcher",
                                            commands)
        common_qt_lib.set_default_window_icon(message_box)
        message_box.setTextInteractionFlags(Qt.TextSelectableByMouse)
        QApplication.clipboard().setText(commands)
        await async_dialog.execute_dialog(message_box)

    @asyncSlot()
    async def _export_iso(self):
        layout = self.layout_description
        has_spoiler = layout.permalink.spoiler
        options = self._options

        if layout.permalink.get_preset(self.current_player_index).configuration.game == RandovaniaGame.PRIME3:
            return await self._show_dialog_for_prime3_layout()

        if not options.is_alert_displayed(InfoAlert.FAQ):
            await async_dialog.message_box(self, QtWidgets.QMessageBox.Icon.Information, "FAQ",
                                           "Have you read the Randovania FAQ?\n"
                                           "It can be found in the main Randovania window → Help → FAQ")
            options.mark_alert_as_displayed(InfoAlert.FAQ)

        game = layout.permalink.get_preset(self.current_player_index).configuration.game
        dialog = GameInputDialog(options, "{} Randomizer - {}.iso".format(game.short_name,
                                                                          layout.shareable_word_hash), has_spoiler)
        result = await async_dialog.execute_dialog(dialog)

        if result != QDialog.Accepted:
            return

        if simplified_patcher.export_busy:
            return await async_dialog.message_box(
                self, QtWidgets.QMessageBox.Critical,
                "Can't save ISO",
                "Error: Unable to save multiple ISOs at the same time,"
                "another window is saving an ISO right now.")

        input_file = dialog.input_file
        output_file = dialog.output_file
        auto_save_spoiler = dialog.auto_save_spoiler
        player_index = self.current_player_index
        player_names = self._player_names

        with options:
            options.output_directory = output_file.parent
            options.auto_save_spoiler = auto_save_spoiler

        def work(progress_update: ProgressUpdateCallable):
            num_updaters = 2
            if input_file is not None:
                num_updaters += 1
            updaters = status_update_lib.split_progress_update(progress_update, num_updaters)

            if input_file is not None:
                simplified_patcher.unpack_iso(input_iso=input_file,
                                              options=options,
                                              progress_update=updaters[0])

            # Apply Layout
            simplified_patcher.apply_layout(layout=layout,
                                            players_config=PlayersConfiguration(
                                                player_index=player_index,
                                                player_names=player_names,
                                            ),
                                            options=options,
                                            progress_update=updaters[-2])

            # Pack ISO
            simplified_patcher.pack_iso(output_iso=output_file,
                                        options=options,
                                        progress_update=updaters[-1])
            if has_spoiler and auto_save_spoiler:
                layout.save_to_file(output_file.with_suffix(f".{LayoutDescription.file_extension()}"))

            progress_update(f"Finished!", 1)

        try:
            await self.run_in_background_async(work, "Exporting...")
        except Exception as e:
            logging.exception("Unable to export ISO")

            self.progress_update_signal.emit(f"Unable to export ISO: {e}", None)
            box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Critical,
                                        "Unable to export ISO",
                                        str(e),
                                        QtWidgets.QMessageBox.Ok)
            common_qt_lib.set_default_window_icon(box)
            if e.__traceback__ is not None:
                box.setDetailedText("".join(traceback.format_tb(e.__traceback__)))
            await async_dialog.execute_dialog(box)

    def _open_map_tracker(self):
        current_preset = self.layout_description.permalink.presets[self.current_player_index]
        self._window_manager.open_map_tracker(current_preset.configuration)

    # Layout Visualization
    def _create_pickup_spoilers(self, game_description: GameDescription):
        if game_description == self._pickup_spoiler_current_game:
            return

        for groups in self._pickup_spoiler_world_to_group.values():
            groups.deleteLater()

        self._pickup_spoiler_current_game = game_description
        self.pickup_spoiler_show_all_button.currently_show_all = True
        self.pickup_spoiler_buttons.clear()

        self._pickup_spoiler_world_to_group = {}
        nodes_in_world = collections.defaultdict(list)

        for world, area, node in game_description.world_list.all_worlds_areas_nodes:
            if isinstance(node, PickupNode):
                world_name = world.correct_name(area.in_dark_aether)
                nodes_in_world[world_name].append((f"{area.name} - {node.name}", node.pickup_index))
                continue

        for world_name in sorted(nodes_in_world.keys()):
            group_box = QGroupBox(self.pickup_spoiler_scroll_contents)
            group_box.setTitle(world_name)
            vertical_layout = QVBoxLayout(group_box)
            vertical_layout.setContentsMargins(8, 4, 8, 4)
            vertical_layout.setSpacing(2)
            group_box.vertical_layout = vertical_layout

            vertical_layout.horizontal_layouts = []
            self._pickup_spoiler_world_to_group[world_name] = group_box
            self.pickup_spoiler_scroll_content_layout.addWidget(group_box)

            for area_name, pickup_index in sorted(nodes_in_world[world_name], key=lambda it: it[0]):
                horizontal_layout = QHBoxLayout()
                horizontal_layout.setSpacing(2)

                label = QLabel(group_box)
                label.setText(area_name)
                horizontal_layout.addWidget(label)
                horizontal_layout.label = label

                push_button = QPushButton(group_box)
                push_button.setFlat(True)
                push_button.setText("Hidden")
                push_button.item_is_hidden = True
                push_button.pickup_index = pickup_index
                push_button.clicked.connect(partial(self._toggle_pickup_spoiler, push_button))
                push_button.item_name = "Nothing was Set, ohno"
                push_button.row = horizontal_layout
                horizontal_layout.addWidget(push_button)
                horizontal_layout.button = push_button
                self.pickup_spoiler_buttons.append(push_button)

                group_box.vertical_layout.addLayout(horizontal_layout)
                group_box.vertical_layout.horizontal_layouts.append(horizontal_layout)

    def create_history_item(self, node):
        button = QRadioButton(self.layout_history_content)
        button.toggled.connect(self.on_select_node)
        button.setText(node)
        self.layout_history_content_layout.addWidget(button)
        self._history_items.append(button)
        return button

    def update_layout_description(self, description: LayoutDescription):
        self.layout_description = description
        self.layout_info_tab.show()

        self.setWindowTitle(f"Game Details: {description.shareable_word_hash}")
        self.export_log_button.setEnabled(description.permalink.spoiler)

        self._player_names = {
            i: f"Player {i + 1}"
            for i in range(description.permalink.player_count)
        }

        self.export_iso_button.setEnabled(description.permalink.player_count == 1)
        if description.permalink.player_count > 1:
            self.export_iso_button.setToolTip("Multiworld games can only be exported from a game session")
        else:
            self.export_iso_button.setToolTip("")

        self.customize_user_preferences_button.setVisible(description.permalink.player_count == 1)

        self.player_index_combo.clear()
        for i in range(description.permalink.player_count):
            self.player_index_combo.addItem(self._player_names[i], i)
        self.player_index_combo.setCurrentIndex(0)
        self.player_index_combo.setVisible(description.permalink.player_count > 1)

        self._update_current_player()

    def _update_current_player(self):
        description = self.layout_description
        current_player = self.current_player_index
        preset = description.permalink.get_preset(current_player)

        self.permalink_edit.setText(description.permalink.as_base64_str)
        title_text = """
        <p>
            Seed Hash: {description.shareable_word_hash} ({description.shareable_hash})<br/>
            Preset Name: {preset.name}
        </p>
        """.format(description=description, preset=preset)
        self.layout_title_label.setText(title_text)

        categories = list(preset_describer.describe(preset))
        self.layout_description_left_label.setText(preset_describer.merge_categories(categories[::2]))
        self.layout_description_right_label.setText(preset_describer.merge_categories(categories[1::2]))

        # Game Spoiler
        has_spoiler = description.permalink.spoiler
        self.pickup_tab.setEnabled(has_spoiler)
        patches = description.all_patches[current_player]

        if has_spoiler:
            pickup_names = {
                pickup.pickup.name
                for pickup in patches.pickup_assignment.values()
            }
            game_description = data_reader.decode_data(preset.configuration.game_data)
            self._create_pickup_spoilers(game_description)
            starting_area = game_description.world_list.area_by_area_location(patches.starting_location)

            extra_items = patcher_file.additional_starting_items(preset.configuration,
                                                                 game_description.resource_database,
                                                                 patches.starting_items)

            self.spoiler_starting_location_label.setText("Starting Location: {}".format(
                game_description.world_list.area_name(starting_area)
            ))
            self.spoiler_starting_items_label.setText("Random Starting Items: {}".format(
                ", ".join(extra_items)
                if extra_items else "None"
            ))
            self._update_show_all_button_state()

        else:
            pickup_names = {}
            self.layout_info_tab.removeTab(self.layout_info_tab.indexOf(self.pickup_tab))
            self.spoiler_starting_location_label.setText("Starting Location")
            self.spoiler_starting_items_label.setText("Random Starting Items")

        self.pickup_spoiler_pickup_combobox.clear()
        self.pickup_spoiler_pickup_combobox.addItem("None")
        for pickup_name in sorted(pickup_names):
            self.pickup_spoiler_pickup_combobox.addItem(pickup_name)

        for pickup_button in self.pickup_spoiler_buttons:
            pickup_target = patches.pickup_assignment.get(pickup_button.pickup_index)

            pickup_button.target_player = None
            if pickup_target is not None:
                pickup_button.item_name = pickup_target.pickup.name if has_spoiler else "????"
                if has_spoiler and description.permalink.player_count > 1:
                    pickup_button.target_player = pickup_target.player
                    pickup_button.player_names = self._player_names
            else:
                pickup_button.item_name = "Nothing"

            if not pickup_button.item_is_hidden:
                pickup_button.setText(pickup_button.item_name)

    def _toggle_pickup_spoiler(self, button):
        if button.item_is_hidden:
            _show_pickup_spoiler(button)
        else:
            _hide_pickup_spoiler(button)
        self._update_show_all_button_state()

    def _toggle_show_all_pickup_spoiler(self):
        if self.pickup_spoiler_show_all_button.currently_show_all:
            action = _show_pickup_spoiler
        else:
            action = _hide_pickup_spoiler

        for button in self.pickup_spoiler_buttons:
            action(button)

        self._update_show_all_button_state()

    def _update_show_all_button_state(self):
        self.pickup_spoiler_show_all_button.currently_show_all = all(
            button.item_is_hidden for button in self.pickup_spoiler_buttons
        )
        if self.pickup_spoiler_show_all_button.currently_show_all:
            self.pickup_spoiler_show_all_button.setText(QtCore.QCoreApplication.translate("HistoryWindow", "Show All"))
        else:
            self.pickup_spoiler_show_all_button.setText(QtCore.QCoreApplication.translate("HistoryWindow", "Hide All"))

    def _on_change_pickup_filter(self, text):
        for button in self.pickup_spoiler_buttons:
            visible = text == "None" or text == button.item_name
            button.setVisible(visible)
            button.row.label.setVisible(visible)

    def _open_user_preferences_dialog(self):
        dialog = EchoesUserPreferencesDialog(self, self._options.cosmetic_patches)
        result = dialog.exec_()
        if result == QDialog.Accepted:
            with self._options as options:
                options.cosmetic_patches = dialog.cosmetic_patches

    def enable_buttons_with_background_tasks(self, value: bool):
        self.stop_background_process_button.setEnabled(not value)
        self.export_iso_button.setEnabled(value)
        simplified_patcher.export_busy = not value

    def update_progress(self, message: str, percentage: int):
        self.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0

        if percentage is None:
            percentage = self._last_percentage
        else:
            self._last_percentage = percentage
        if percentage >= 0:
            self.progress_bar.setRange(0, 100)
            self.progress_bar.setValue(percentage)
        else:
            self.progress_bar.setRange(0, 0)



import base64
import hashlib
import json
import logging
import typing
from typing import Optional, List, Tuple

import flask_socketio
import peewee

from randovania.bitpacking import bitpacking
from randovania.game_description import data_reader
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games.game import RandovaniaGame
from randovania.games.prime import patcher_file
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.preset_migration import VersionedPreset
from randovania.network_common.admin_actions import SessionAdminGlobalAction, SessionAdminUserAction
from randovania.network_common.error import WrongPassword, \
    NotAuthorizedForAction, InvalidAction
from randovania.network_common.pickup_serializer import BitPackPickupEntry
from randovania.network_common.session_state import GameSessionState
from randovania.server import database
from randovania.server.database import GameSession, GameSessionMembership, GameSessionTeamAction, \
    GameSessionPreset
from randovania.server.lib import logger
from randovania.server.server_app import ServerApp


def list_game_sessions(sio: ServerApp):
    return [
        session.create_list_entry()
        for session in GameSession.select()
    ]


def create_game_session(sio: ServerApp, session_name: str):
    current_user = sio.get_current_user()

    with database.db.atomic():
        new_session = GameSession.create(
            name=session_name,
            password=None,
            creator=current_user,
        )
        GameSessionPreset.create(session=new_session, row=0,
                                 preset=json.dumps(PresetManager(None).default_preset.as_json))
        membership = GameSessionMembership.create(
            user=sio.get_current_user(), session=new_session,
            row=0, admin=True, connection_state="Online, Unknown")

    sio.join_game_session(membership)
    return new_session.create_session_entry()


def join_game_session(sio: ServerApp, session_id: int, password: Optional[str]):
    session = GameSession.get_by_id(session_id)

    if session.password is not None:
        if password is None or _hash_password(password) != session.password:
            raise WrongPassword()
    elif password is not None:
        raise WrongPassword()

    membership = GameSessionMembership.get_or_create(user=sio.get_current_user(), session=session,
                                                     defaults={"row": None, "admin": False,
                                                               "connection_state": "Online, Unknown"})[0]

    _emit_session_update(session)
    sio.join_game_session(membership)

    return session.create_session_entry()


def disconnect_game_session(sio: ServerApp, session_id: int):
    current_user = sio.get_current_user()
    try:
        current_membership = GameSessionMembership.get_by_ids(current_user.id, session_id)
        current_membership.connection_state = "Offline"
        current_membership.save()
        _emit_session_update(current_membership.session)
    except peewee.DoesNotExist:
        pass
    sio.leave_game_session()


def _verify_has_admin(sio: ServerApp, session_id: int, admin_user_id: Optional[int],
                      *, allow_when_no_admins: bool = False) -> None:
    """
    Checks if the logged user can do admin operations to the given session,
    :param session_id: The GameSessions id
    :param admin_user_id: An user id that is exceptionally authorized for this
    :param allow_when_no_admins: This action is authorized for non-admins if there are no admins.
    :return:
    """
    current_user = sio.get_current_user()
    try:
        current_membership = GameSessionMembership.get_by_ids(current_user.id, session_id)
    except peewee.DoesNotExist:
        raise NotAuthorizedForAction()

    if not (current_membership.admin or (admin_user_id is not None and current_user.id == admin_user_id)):
        if allow_when_no_admins and GameSessionMembership.select().where(
                GameSessionMembership.session == session_id, GameSessionMembership.admin == True).count() == 0:
            return
        raise NotAuthorizedForAction()


def _verify_in_setup(session: GameSession):
    if session.state != GameSessionState.SETUP:
        raise InvalidAction("Session is not in setup")


def _verify_no_layout_description(session: GameSession):
    if session.layout_description_json is not None:
        raise InvalidAction("Session has a generated game")


def _get_preset(preset_json: dict) -> VersionedPreset:
    try:
        preset = VersionedPreset(preset_json)
        preset.get_preset()  # test if valid
        return preset
    except Exception as e:
        raise InvalidAction(f"invalid preset: {e}")


def _emit_session_update(session: GameSession):
    flask_socketio.emit("game_session_update", session.create_session_entry(), room=f"game-session-{session.id}")


def game_session_request_update(sio: ServerApp, session_id):
    session: database.GameSession = database.GameSession.get_by_id(session_id)
    return session.create_session_entry()


def _create_row(sio: ServerApp, session: GameSession, preset_json: dict):
    _verify_has_admin(sio, session.id, None)
    _verify_in_setup(session)
    _verify_no_layout_description(session)
    preset = _get_preset(preset_json)

    new_row_id = session.num_rows
    with database.db.atomic():
        logger().info(f"Session {session.id}: Creating row {new_row_id}.")
        GameSessionPreset.create(session=session, row=new_row_id,
                                 preset=json.dumps(preset.as_json))


def _change_row(sio: ServerApp, session: GameSession, arg: Tuple[int, dict]):
    if len(arg) != 2:
        raise InvalidAction("Missing arguments.")
    row_id, preset_json = arg
    _verify_has_admin(sio, session.id, sio.get_current_user().id)
    _verify_in_setup(session)
    _verify_no_layout_description(session)
    preset = _get_preset(preset_json)
    if preset.game != RandovaniaGame.PRIME2:
        raise InvalidAction("Only Prime 2 presets allowed.")

    try:
        with database.db.atomic():
            preset_row = GameSessionPreset.get(GameSessionPreset.session == session,
                                               GameSessionPreset.row == row_id)
            preset_row.preset = json.dumps(preset.as_json)
            logger().info(f"Session {session.id}: Changing row {row_id}.")
            preset_row.save()

    except peewee.DoesNotExist:
        raise InvalidAction(f"invalid row: {row_id}")


def _delete_row(sio: ServerApp, session: GameSession, row_id: int):
    _verify_has_admin(sio, session.id, None)
    _verify_in_setup(session)
    _verify_no_layout_description(session)

    if session.num_rows < 2:
        raise InvalidAction("Can't delete row when there's only one")

    if row_id != session.num_rows - 1:
        raise InvalidAction(f"Can only delete the last row")

    with database.db.atomic():
        logger().info(f"Session {session.id}: Deleting {row_id}.")
        GameSessionPreset.delete().where(GameSessionPreset.session == session,
                                         GameSessionPreset.row == row_id).execute()
        GameSessionMembership.update(row=None).where(
            GameSessionMembership.session == session.id,
            GameSessionMembership.row == row_id,
        ).execute()


def _update_layout_generation(sio: ServerApp, session: GameSession, active: bool):
    _verify_has_admin(sio, session.id, None)
    _verify_in_setup(session)

    if active:
        if session.generation_in_progress is None:
            session.generation_in_progress = sio.get_current_user()
        else:
            raise InvalidAction(f"Generation already in progress by {session.generation_in_progress.name}.")
    else:
        session.generation_in_progress = None

    logger().info(f"Session {session.id}: Making generation in progress to {session.generation_in_progress}.")
    session.save()


def _change_layout_description(sio: ServerApp, session: GameSession, description_json: Optional[dict]):
    _verify_has_admin(sio, session.id, None)
    _verify_in_setup(session)
    rows_to_update = []

    if description_json is None:
        description = None
    else:
        if session.generation_in_progress != sio.get_current_user():
            if session.generation_in_progress is None:
                raise InvalidAction(f"Not waiting for a layout.")
            else:
                raise InvalidAction(f"Waiting for a layout from {session.generation_in_progress.name}.")

        _verify_no_layout_description(session)
        description = LayoutDescription.from_json_dict(description_json)
        permalink = description.permalink
        if permalink.player_count != session.num_rows:
            raise InvalidAction(f"Description is for a {permalink.player_count} players,"
                                f" while the session is for {session.num_rows}.")

        for permalink_preset, preset_row in zip(permalink.presets.values(), session.presets):
            preset_row = typing.cast(GameSessionPreset, preset_row)
            if _get_preset(json.loads(preset_row.preset)).get_preset() != permalink_preset:
                preset = VersionedPreset.with_preset(permalink_preset)
                if preset.game != RandovaniaGame.PRIME2:
                    raise InvalidAction("Only Prime 2 presets allowed.")
                preset_row.preset = json.dumps(preset.as_json)
                rows_to_update.append(preset_row)

    with database.db.atomic():
        for preset_row in rows_to_update:
            preset_row.save()

        session.generation_in_progress = None
        session.layout_description = description
        session.save()


def _download_layout_description(sio: ServerApp, session: GameSession):
    try:
        # You must be a session member to do get the spoiler
        GameSessionMembership.get_by_ids(sio.get_current_user().id, session.id)
    except peewee.DoesNotExist:
        raise NotAuthorizedForAction()

    if session.layout_description_json is None:
        raise InvalidAction("Session does not contain a game")

    if not session.layout_description.permalink.spoiler:
        raise InvalidAction("Session does not contain a spoiler")

    return session.layout_description_json


def _start_session(sio: ServerApp, session: GameSession):
    _verify_has_admin(sio, session.id, None)
    _verify_in_setup(session)
    if session.layout_description_json is None:
        raise InvalidAction("Unable to start session, no game is available.")

    num_players = GameSessionMembership.select().where(GameSessionMembership.session == session,
                                                       GameSessionMembership.row != None).count()
    expected_players = session.num_rows
    if num_players != expected_players:
        raise InvalidAction(f"Unable to start session, there are {num_players} but expected {expected_players} "
                            f"({session.num_rows} x {session.num_teams}).")

    session.state = GameSessionState.IN_PROGRESS
    logger().info(f"Session {session.id}: Starting session.")
    session.save()


def _finish_session(sio: ServerApp, session: GameSession):
    _verify_has_admin(sio, session.id, None)
    if session.state != GameSessionState.IN_PROGRESS:
        raise InvalidAction("Session is not in progress")

    session.state = GameSessionState.FINISHED
    logger().info(f"Session {session.id}: Finishing session.")
    session.save()


def _reset_session(sio: ServerApp, session: GameSession):
    raise InvalidAction("Restart session is not yet implemented.")


def _hash_password(password: str) -> str:
    return hashlib.blake2s(password.encode("utf-8")).hexdigest()


def _change_password(sio: ServerApp, session: GameSession, password: str):
    _verify_has_admin(sio, session.id, None)

    session.password = _hash_password(password)
    logger().info(f"Session {session.id}: Changing password.")
    session.save()


def game_session_admin_session(sio: ServerApp, session_id: int, action: str, arg):
    action: SessionAdminGlobalAction = SessionAdminGlobalAction(action)
    session: database.GameSession = database.GameSession.get_by_id(session_id)

    if action == SessionAdminGlobalAction.CREATE_ROW:
        _create_row(sio, session, arg)

    elif action == SessionAdminGlobalAction.CHANGE_ROW:
        _change_row(sio, session, arg)

    elif action == SessionAdminGlobalAction.DELETE_ROW:
        _delete_row(sio, session, arg)

    elif action == SessionAdminGlobalAction.UPDATE_LAYOUT_GENERATION:
        _update_layout_generation(sio, session, arg)

    elif action == SessionAdminGlobalAction.CHANGE_LAYOUT_DESCRIPTION:
        _change_layout_description(sio, session, arg)

    elif action == SessionAdminGlobalAction.DOWNLOAD_LAYOUT_DESCRIPTION:
        return _download_layout_description(sio, session)

    elif action == SessionAdminGlobalAction.START_SESSION:
        _start_session(sio, session)

    elif action == SessionAdminGlobalAction.FINISH_SESSION:
        _finish_session(sio, session)

    elif action == SessionAdminGlobalAction.RESET_SESSION:
        _reset_session(sio, session)

    elif action == SessionAdminGlobalAction.CHANGE_PASSWORD:
        _change_password(sio, session, arg)

    elif action == SessionAdminGlobalAction.DELETE_SESSION:
        logger().info(f"Session {session.id}: Deleting session.")
        session.delete_instance(recursive=True)

    _emit_session_update(session)


def _find_empty_row(session: GameSession) -> int:
    possible_rows = set(range(session.num_rows))
    for member in GameSessionMembership.non_observer_members(session):
        possible_rows.remove(member.row)

    for empty_row in sorted(possible_rows):
        return empty_row
    raise InvalidAction("Session is full")


def game_session_admin_player(sio: ServerApp, session_id: int, user_id: int, action: str, arg):
    _verify_has_admin(sio, session_id, user_id)
    action: SessionAdminUserAction = SessionAdminUserAction(action)

    session: GameSession = database.GameSession.get_by_id(session_id)
    membership = GameSessionMembership.get_by_ids(user_id, session_id)

    if action == SessionAdminUserAction.KICK:
        membership.delete_instance()
        if not list(session.players):
            session.delete_instance(recursive=True)
            logger().info(f"Session {session_id}. Kicking user {user_id} and deleting session.")
        else:
            logger().info(f"Session {session_id}. Kicking user {user_id}.")

    elif action == SessionAdminUserAction.MOVE:
        offset: int = arg
        if membership.is_observer is None:
            raise InvalidAction("Player is an observer")

        new_row = membership.row + offset
        if new_row < 0:
            raise InvalidAction("New position is negative")
        if new_row >= session.num_rows:
            raise InvalidAction("New position is beyond num of rows")

        team_members = [None] * session.num_rows
        for member in GameSessionMembership.non_observer_members(session):
            team_members[member.row] = member

        while (0 <= new_row < session.num_rows) and team_members[new_row] is not None:
            new_row += offset

        if new_row < 0 or new_row >= session.num_rows:
            raise InvalidAction("No empty slots found in this direction")

        with database.db.atomic():
            logger().info(f"Session {session_id}, User {user_id}. "
                          f"Performing {action}, new row is {new_row}, from {membership.row}.")
            membership.row = new_row
            membership.save()

    elif action == SessionAdminUserAction.SWITCH_IS_OBSERVER:
        if membership.is_observer:
            membership.row = _find_empty_row(session)
        else:
            membership.row = None
        logger().info(f"Session {session_id}, User {user_id}. Performing {action}, new row is {membership.row}.")
        membership.save()

    elif action == SessionAdminUserAction.SWITCH_ADMIN:
        # Must be admin for this
        _verify_has_admin(sio, session_id, None, allow_when_no_admins=True)
        num_admins = GameSessionMembership.select().where(GameSessionMembership.session == session_id,
                                                          GameSessionMembership.admin == True).count()

        if membership.admin and num_admins <= 1:
            raise InvalidAction("can't demote the only admin")

        membership.admin = not membership.admin
        logger().info(f"Session {session_id}, User {user_id}. Performing {action}, new status is {membership.admin}.")
        membership.save()

    elif action == SessionAdminUserAction.CREATE_PATCHER_FILE:
        cosmetic_patches = CosmeticPatches.from_json_dict(arg)
        player_names = {i: f"Player {i + 1}" for i in range(session.num_rows)}

        for member in GameSessionMembership.non_observer_members(session):
            player_names[member.row] = member.effective_name

        players_config = PlayersConfiguration(
            player_index=membership.row,
            player_names=player_names,
        )
        return patcher_file.create_patcher_file(session.layout_description,
                                                players_config,
                                                cosmetic_patches)

    elif action == SessionAdminUserAction.ABANDON:
        # FIXME
        raise InvalidAction("Abandon is NYI")

    _emit_session_update(session)


def _query_for_actions(membership: GameSessionMembership) -> peewee.ModelSelect:
    return GameSessionTeamAction.select().where(
        GameSessionTeamAction.provider_row != membership.row,
        GameSessionTeamAction.session == membership.session,
        GameSessionTeamAction.receiver_row == membership.row,
    ).order_by(GameSessionTeamAction.time.asc())


def _base64_encode_pickup(pickup: PickupEntry, resource_database: ResourceDatabase) -> str:
    encoded_pickup = bitpacking.pack_value(BitPackPickupEntry(pickup, resource_database))
    return base64.b85encode(encoded_pickup).decode("utf-8")


def _collect_location(session: GameSession, membership: GameSessionMembership,
                      description: LayoutDescription,
                      pickup_location: int) -> Optional[int]:
    """
    Collects the pickup in the given location. Returns
    :param session:
    :param membership:
    :param description:
    :param pickup_location:
    :return: The rewarded player if some player must be updated of the fact.
    """
    player_row: int = membership.row
    pickup_target = _get_pickup_target(description, player_row, pickup_location)

    def log(msg):
        logger().info(f"Session {session.id}, Row {membership.row} found item at {pickup_location}. {msg}")

    if pickup_target is None:
        log(f"It's an ETM.")
        return None

    if pickup_target.player == membership.row:
        log(f"It's a {pickup_target.pickup.name} for themselves.")
        return None

    try:
        GameSessionTeamAction.create(
            session=session,
            provider_row=membership.row,
            provider_location_index=pickup_location,
            receiver_row=pickup_target.player,
        )
    except peewee.IntegrityError:
        # Already exists and it's for another player, no inventory update needed
        log(f"It's a {pickup_target.pickup.name} for {pickup_target.player}, but it was already collected.")
        return None

    log(f"It's a {pickup_target.pickup.name} for {pickup_target.player}.")
    return pickup_target.player


def game_session_collect_locations(sio: ServerApp, session_id: int, pickup_locations: Tuple[int, ...]):
    current_user = sio.get_current_user()
    session: GameSession = database.GameSession.get_by_id(session_id)
    membership = GameSessionMembership.get_by_ids(current_user.id, session_id)

    if session.state != GameSessionState.IN_PROGRESS:
        raise InvalidAction("Unable to collect locations of sessions that aren't in progress")

    if membership.is_observer:
        raise InvalidAction("Observers can't collect locations")

    description = session.layout_description

    receiver_players = set()
    for location in pickup_locations:
        receiver_player = _collect_location(session, membership, description, location)
        if receiver_player is not None:
            receiver_players.add(receiver_player)

    if not receiver_players:
        return

    for receiver_player in receiver_players:
        try:
            receiver_membership = GameSessionMembership.get_by_session_position(session, row=receiver_player)
            flask_socketio.emit(
                "game_has_update",
                {
                    "session": session_id,
                    "row": receiver_player,
                },
                room=f"game-session-{receiver_membership.session.id}-{receiver_membership.user.id}")
        except peewee.DoesNotExist:
            pass
    _emit_session_update(session)


def _get_resource_database(description: LayoutDescription, player: int) -> ResourceDatabase:
    game_data = description.permalink.get_preset(player).configuration.game_data
    return data_reader.read_resource_database(game_data["resource_database"])


def _get_pickup_target(description: LayoutDescription, provider: int, location: int) -> Optional[PickupTarget]:
    pickup_assignment = description.all_patches[provider].pickup_assignment
    return pickup_assignment.get(PickupIndex(location))


def game_session_request_pickups(sio: ServerApp, session_id: int):
    current_user = sio.get_current_user()
    your_membership = GameSessionMembership.get_by_ids(current_user.id, session_id)
    session: GameSession = your_membership.session

    if session.state == GameSessionState.SETUP:
        logger().info(f"Session {session_id}, Row {your_membership.row} "
                      f"requested pickups, but session is setup.")
        return []

    if your_membership.is_observer:
        logger().info(f"Session {session_id}, {current_user.name} requested pickups, but is an observer.")
        return []

    description = session.layout_description
    row_to_member_name = {
        member.row: member.effective_name
        for member in GameSessionMembership.non_observer_members(session)
    }

    resource_database = _get_resource_database(description, your_membership.row)

    result = []
    actions: List[GameSessionTeamAction] = list(_query_for_actions(your_membership))
    for action in actions:
        pickup_target = _get_pickup_target(description, action.provider_row, action.provider_location_index)

        if pickup_target is None:
            logging.error(f"Action {action} has a location index with nothing.")
            result.append(None)
        else:
            name = row_to_member_name.get(action.provider_row, f"Player {action.provider_row + 1}")
            result.append({
                "provider_name": name,
                "pickup": _base64_encode_pickup(pickup_target.pickup, resource_database),
            })

    logger().info(f"Session {session_id}, Row {your_membership.row} "
                  f"requested pickups, returning {len(result)} elements.")

    return result


def game_session_self_update(sio: ServerApp, session_id: int, inventory: str, game_connection_state: str):
    current_user = sio.get_current_user()
    membership = GameSessionMembership.get_by_ids(current_user.id, session_id)

    membership.connection_state = f"Online, {game_connection_state}"
    membership.inventory = inventory
    membership.save()
    _emit_session_update(membership.session)


def report_user_disconnected(sio: ServerApp, user_id: int, log):
    memberships: List[GameSessionMembership] = list(GameSessionMembership.select().where(
        GameSessionMembership.user == user_id))

    log.info(f"User {user_id} is disconnected, disconnecting from sessions: {memberships}")
    sessions_to_update = []

    for membership in memberships:
        if membership.connection_state != "Offline":
            membership.connection_state = "Offline"
            sessions_to_update.append(membership.session)
            membership.save()

    for session in sessions_to_update:
        _emit_session_update(session)


def setup_app(sio: ServerApp):
    sio.on("list_game_sessions", list_game_sessions)
    sio.on("create_game_session", create_game_session)
    sio.on("join_game_session", join_game_session)
    sio.on("disconnect_game_session", disconnect_game_session)
    sio.on("game_session_request_update", game_session_request_update)
    sio.on("game_session_admin_session", game_session_admin_session)
    sio.on("game_session_admin_player", game_session_admin_player)
    sio.on("game_session_collect_locations", game_session_collect_locations)
    sio.on("game_session_request_pickups", game_session_request_pickups)
    sio.on("game_session_self_update", game_session_self_update)



import pytest

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import ConditionalResources, ResourceConversion, PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import add_resource_gain_to_current_resources, \
    add_resources_into_another, convert_resource_gain_to_current_resources


@pytest.mark.parametrize(["a", "b", "result"], [
    ({"a": 5}, {"b": 6}, {"a": 5, "b": 6}),
    ({"a": 5}, {"a": 6}, {"a": 11}),
])
def test_add_resources_into_another(a, b, result):
    add_resources_into_another(a, b)
    assert a == result


def test_pickup_index_equality():
    assert PickupIndex(1) == PickupIndex(1)


def test_pickup_index_has():
    d = {PickupIndex(1): True}
    assert PickupIndex(1) in d


def test_add_resource_gain_to_current_resources_convert():
    # Setup
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    pickup = PickupEntry(
        name="P1", model_index=1, item_category=ItemCategory.SUIT, broad_category=ItemCategory.LIFE_SUPPORT,
        resources=(ConditionalResources(None, None, ()),),
        convert_resources=(ResourceConversion(resource_a, resource_b),)
    )
    current_resources = {
        resource_a: 5
    }

    # Run
    add_resource_gain_to_current_resources(pickup.resource_gain(current_resources), current_resources)

    # Assert
    assert current_resources == {
        resource_a: 0,
        resource_b: 5
    }


@pytest.mark.parametrize(["resource_gain", "expected"], [
    ([], {}),
    ([("a", 5), ("b", 6)], {"a": 5, "b": 6}),
    ([("a", 5), ("a", 6)], {"a": 11}),
    ([("a", 5), ("a", -5)], {"a": 0}),
])
def test_convert_resource_gain_to_current_resources(resource_gain, expected):
    # Setup
    # Run
    result = convert_resource_gain_to_current_resources(resource_gain)

    # Assert
    assert result == expected



from unittest.mock import MagicMock

import pytest

from randovania.dol_patching import dol_version
from randovania.games.game import RandovaniaGame

test_versions = [
    dol_version.DolVersion(
        game=RandovaniaGame.PRIME2,
        description="Gamecube NTSC",
        build_string_address=0x803ac3b0,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32",
        sda2_base=0x804223c0,
    ),
    dol_version.DolVersion(
        game=RandovaniaGame.PRIME2,
        description="Gamecube PAL",
        build_string_address=0x803ad710,
        build_string=b"!#$MetroidBuildInfo!#$Build v1.035 10/27/2004 19:48:17",
        sda2_base=0x804223c0,
    ),
]


@pytest.mark.parametrize("version", test_versions)
def test_read_binary_version(version):
    dol_file = MagicMock()
    dol_file.read.return_value = version.build_string

    # Run
    result = dol_version.find_version_for_dol(dol_file, test_versions)

    # Assert
    dol_file.set_editable.assert_called_once_with(False)
    assert result == version



from typing import Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackFloat, BitPackDecoder
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import PickupEntry, MAXIMUM_PICKUP_CONDITIONAL_RESOURCES, \
    MAXIMUM_PICKUP_RESOURCES, MAXIMUM_PICKUP_CONVERSION, ConditionalResources, ResourceConversion
from randovania.game_description.resources.resource_database import ResourceDatabase

_PROBABILITY_OFFSET_META = {
    "min": -3,
    "max": 3,
    "precision": 2.0,
    "if_different": 0.0,
}
_PROBABILITY_MULTIPLIER_META = {
    "min": 0,
    "max": 8,
    "precision": 2.0,
    "if_different": 1.0,
}


class BitPackPickupEntry:
    value: PickupEntry
    database: ResourceDatabase

    def __init__(self, value: PickupEntry, database: ResourceDatabase):
        self.value = value
        self.database = database

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        items = self.database.item

        yield from bitpacking.encode_string(self.value.name)
        yield self.value.model_index, 255
        yield from BitPackFloat(self.value.probability_offset).bit_pack_encode(_PROBABILITY_OFFSET_META)
        yield from BitPackFloat(self.value.probability_multiplier).bit_pack_encode(_PROBABILITY_MULTIPLIER_META)
        yield from self.value.item_category.bit_pack_encode({})
        yield from self.value.broad_category.bit_pack_encode({})
        yield len(self.value.resources) - 1, MAXIMUM_PICKUP_CONDITIONAL_RESOURCES

        for i, conditional in enumerate(self.value.resources):
            if i > 0:
                yield from bitpacking.pack_array_element(conditional.item, items)

            yield len(conditional.resources), MAXIMUM_PICKUP_RESOURCES + 1
            for resource, quantity in conditional.resources:
                yield from bitpacking.pack_array_element(resource, items)
                yield quantity, 255

            yield from bitpacking.encode_bool(conditional.name is not None)
            if conditional.name is not None:
                yield from bitpacking.encode_string(conditional.name)

        yield len(self.value.convert_resources), MAXIMUM_PICKUP_CONVERSION + 1
        for conversion in self.value.convert_resources:
            yield from bitpacking.pack_array_element(conversion.source, items)
            yield from bitpacking.pack_array_element(conversion.target, items)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, database: ResourceDatabase) -> PickupEntry:
        name = bitpacking.decode_string(decoder)
        model_index = decoder.decode_single(255)
        probability_offset = BitPackFloat.bit_pack_unpack(decoder, _PROBABILITY_OFFSET_META)
        probability_multiplier = BitPackFloat.bit_pack_unpack(decoder, _PROBABILITY_MULTIPLIER_META)
        item_category = ItemCategory.bit_pack_unpack(decoder, {})
        broad_category = ItemCategory.bit_pack_unpack(decoder, {})
        num_conditional = decoder.decode_single(MAXIMUM_PICKUP_CONDITIONAL_RESOURCES) + 1

        conditional_resources = []
        for i in range(num_conditional):
            item_name = None  # TODO: get the first resource name
            if i > 0:
                item_dependency = decoder.decode_element(database.item)
            else:
                item_dependency = None

            resources = []
            for _ in range(decoder.decode_single(MAXIMUM_PICKUP_RESOURCES + 1)):
                resource = decoder.decode_element(database.item)
                quantity = decoder.decode_single(255)
                resources.append((resource, quantity))

            if bitpacking.decode_bool(decoder):
                item_name = bitpacking.decode_string(decoder)

            conditional_resources.append(ConditionalResources(
                name=item_name,
                item=item_dependency,
                resources=tuple(resources),
            ))

        num_convert = decoder.decode_single(MAXIMUM_PICKUP_CONVERSION + 1)
        convert_resources = []
        for i in range(num_convert):
            source = decoder.decode_element(database.item)
            target = decoder.decode_element(database.item)
            convert_resources.append(ResourceConversion(source, target))

        return PickupEntry(
            name=name,
            model_index=model_index,
            item_category=item_category,
            broad_category=broad_category,
            resources=tuple(conditional_resources),
            convert_resources=tuple(convert_resources),
            probability_offset=probability_offset,
            probability_multiplier=probability_multiplier,
        )



from typing import List

import pytest

from randovania.game_description.assignment import PickupTarget
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.prime import echoes_items
from randovania.games.prime.patcher_file_lib import sky_temple_key_hint
from randovania.generator.item_pool import pickup_creator
from randovania.interface_common.players_configuration import PlayersConfiguration


def _create_hint_text(hide_area: bool,
                      key_number: int,
                      key_location: str,
                      ) -> List[str]:
    if hide_area:
        key_location = key_location.split(" - ")[0]

    message = f"&push;&main-color=#FF6705B3;Sky Temple Key {key_number}&pop; is located in &push;&main-color=#FF3333;{key_location}&pop;."
    return [message, "", message]


def make_starting_stk_hint(key_number: int) -> List[str]:
    message = f"&push;&main-color=#FF6705B3;Sky Temple Key {key_number}&pop; has no need to be located."
    return [message, "", message]


def make_useless_stk_hint(key_number: int) -> List[str]:
    message = f"&push;&main-color=#FF6705B3;Sky Temple Key {key_number}&pop; is lost somewhere in Aether."
    return [message, "", message]


@pytest.mark.parametrize("hide_area", [False, True])
def test_create_hints_all_placed(hide_area: bool,
                                 empty_patches, echoes_game_description):
    # Setup
    patches = empty_patches.assign_new_pickups([
        (PickupIndex(17 + key),
         PickupTarget(pickup_creator.create_sky_temple_key(key, echoes_game_description.resource_database), 0))
        for key in range(9)
    ])
    expected = [
        {"asset_id": 0xD97685FE,
         "strings": _create_hint_text(hide_area, 1, "Sky Temple Grounds - Profane Path")},
        {"asset_id": 0x32413EFD,
         "strings": _create_hint_text(hide_area, 2, "Sky Temple Grounds - Phazon Grounds")},
        {"asset_id": 0xDD8355C3,
         "strings": _create_hint_text(hide_area, 3, "Sky Temple Grounds - Ing Reliquary")},
        {"asset_id": 0x3F5F4EBA,
         "strings": _create_hint_text(hide_area, 4, "Great Temple - Transport A Access")},
        {"asset_id": 0xD09D2584,
         "strings": _create_hint_text(hide_area, 5, "Great Temple - Temple Sanctuary")},
        {"asset_id": 0x3BAA9E87,
         "strings": _create_hint_text(hide_area, 6, "Great Temple - Transport B Access")},
        {"asset_id": 0xD468F5B9,
         "strings": _create_hint_text(hide_area, 7, "Great Temple - Main Energy Controller")},
        {"asset_id": 0x2563AE34,
         "strings": _create_hint_text(hide_area, 8, "Great Temple - Main Energy Controller")},
        {"asset_id": 0xCAA1C50A,
         "strings": _create_hint_text(hide_area, 9, "Agon Wastes - Mining Plaza")},
    ]

    # Run
    result = sky_temple_key_hint.create_hints({0: patches}, PlayersConfiguration(0, {0: "you"}),
                                              echoes_game_description.world_list, hide_area)

    # Assert
    assert result == expected


@pytest.mark.parametrize("hide_area", [False, True])
def test_create_hints_all_starting(hide_area: bool,
                                   empty_patches, echoes_game_description):
    # Setup
    patches = empty_patches.assign_extra_initial_items({
        echoes_game_description.resource_database.get_item(echoes_items.SKY_TEMPLE_KEY_ITEMS[key]): 1
        for key in range(9)
    })

    expected = [
        {"asset_id": 0xD97685FE,
         "strings": make_starting_stk_hint(1)},
        {"asset_id": 0x32413EFD,
         "strings": make_starting_stk_hint(2)},
        {"asset_id": 0xDD8355C3,
         "strings": make_starting_stk_hint(3)},
        {"asset_id": 0x3F5F4EBA,
         "strings": make_starting_stk_hint(4)},
        {"asset_id": 0xD09D2584,
         "strings": make_starting_stk_hint(5)},
        {"asset_id": 0x3BAA9E87,
         "strings": make_starting_stk_hint(6)},
        {"asset_id": 0xD468F5B9,
         "strings": make_starting_stk_hint(7)},
        {"asset_id": 0x2563AE34,
         "strings": make_starting_stk_hint(8)},
        {"asset_id": 0xCAA1C50A,
         "strings": make_starting_stk_hint(9)},
    ]

    # Run
    result = sky_temple_key_hint.create_hints({0: patches}, PlayersConfiguration(0, {0: "you"}),
                                              echoes_game_description.world_list, hide_area)

    # Assert
    assert result == expected


def test_hide_hints():
    # Setup
    expected = [
        {"asset_id": 0xD97685FE,
         "strings": make_useless_stk_hint(1)},
        {"asset_id": 0x32413EFD,
         "strings": make_useless_stk_hint(2)},
        {"asset_id": 0xDD8355C3,
         "strings": make_useless_stk_hint(3)},
        {"asset_id": 0x3F5F4EBA,
         "strings": make_useless_stk_hint(4)},
        {"asset_id": 0xD09D2584,
         "strings": make_useless_stk_hint(5)},
        {"asset_id": 0x3BAA9E87,
         "strings": make_useless_stk_hint(6)},
        {"asset_id": 0xD468F5B9,
         "strings": make_useless_stk_hint(7)},
        {"asset_id": 0x2563AE34,
         "strings": make_useless_stk_hint(8)},
        {"asset_id": 0xCAA1C50A,
         "strings": make_useless_stk_hint(9)},
    ]

    # Run
    result = sky_temple_key_hint.hide_hints()

    # Assert
    assert result == expected



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_item_pool.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetItemPool(object):
    def setupUi(self, PresetItemPool):
        if not PresetItemPool.objectName():
            PresetItemPool.setObjectName(u"PresetItemPool")
        PresetItemPool.resize(566, 450)
        PresetItemPool.setMaximumSize(QSize(16777215, 16777215))
        self.centralWidget = QWidget(PresetItemPool)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setFrameShadow(QFrame.Plain)
        self.scroll_area.setLineWidth(0)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 566, 450))
        self.item_pool_layout = QVBoxLayout(self.scroll_area_contents)
        self.item_pool_layout.setSpacing(6)
        self.item_pool_layout.setContentsMargins(11, 11, 11, 11)
        self.item_pool_layout.setObjectName(u"item_pool_layout")
        self.item_pool_layout.setContentsMargins(0, 6, 0, 0)
        self.item_pool_count_label = QLabel(self.scroll_area_contents)
        self.item_pool_count_label.setObjectName(u"item_pool_count_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.item_pool_count_label.sizePolicy().hasHeightForWidth())
        self.item_pool_count_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(12)
        self.item_pool_count_label.setFont(font)
        self.item_pool_count_label.setAlignment(Qt.AlignCenter)

        self.item_pool_layout.addWidget(self.item_pool_count_label)

        self.random_starting_box = QGroupBox(self.scroll_area_contents)
        self.random_starting_box.setObjectName(u"random_starting_box")
        self.gridLayout_2 = QGridLayout(self.random_starting_box)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.maximum_starting_label = QLabel(self.random_starting_box)
        self.maximum_starting_label.setObjectName(u"maximum_starting_label")

        self.gridLayout_2.addWidget(self.maximum_starting_label, 2, 0, 1, 1)

        self.random_starting_label = QLabel(self.random_starting_box)
        self.random_starting_label.setObjectName(u"random_starting_label")
        self.random_starting_label.setWordWrap(True)

        self.gridLayout_2.addWidget(self.random_starting_label, 0, 0, 1, 2)

        self.maximum_starting_spinbox = QSpinBox(self.random_starting_box)
        self.maximum_starting_spinbox.setObjectName(u"maximum_starting_spinbox")
        self.maximum_starting_spinbox.setMaximum(30)

        self.gridLayout_2.addWidget(self.maximum_starting_spinbox, 2, 1, 1, 1)

        self.minimum_starting_label = QLabel(self.random_starting_box)
        self.minimum_starting_label.setObjectName(u"minimum_starting_label")
        self.minimum_starting_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.minimum_starting_label, 1, 0, 1, 1)

        self.minimum_starting_spinbox = QSpinBox(self.random_starting_box)
        self.minimum_starting_spinbox.setObjectName(u"minimum_starting_spinbox")
        self.minimum_starting_spinbox.setMaximum(30)

        self.gridLayout_2.addWidget(self.minimum_starting_spinbox, 1, 1, 1, 1)


        self.item_pool_layout.addWidget(self.random_starting_box)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        PresetItemPool.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetItemPool)

        QMetaObject.connectSlotsByName(PresetItemPool)
    # setupUi

    def retranslateUi(self, PresetItemPool):
        PresetItemPool.setWindowTitle(QCoreApplication.translate("PresetItemPool", u"Item Pool", None))
#if QT_CONFIG(tooltip)
        self.item_pool_count_label.setToolTip(QCoreApplication.translate("PresetItemPool", u"If there are fewer than 119 items, the rest of the item locations will contain Energy Transfer Modules.", None))
#endif // QT_CONFIG(tooltip)
        self.item_pool_count_label.setText(QCoreApplication.translate("PresetItemPool", u"Items in pool: #/119", None))
        self.random_starting_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Random Starting Items", None))
        self.maximum_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"Start with at most this many items:", None))
        self.random_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"<html><head/><body><p>Randovania will add additional starting items if necessary to make the seed possible.<br/>The first value controls how many items are always added.<br/>The second value controls how many items the seed can have before it fails to generate.</p></body></html>", None))
        self.minimum_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"Start with at least this many items:", None))
    # retranslateUi




import copy
from typing import Tuple

from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import PlayerShipNode
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources, \
    add_resource_gain_to_current_resources
from randovania.games.game import RandovaniaGame
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.major_items_configuration import MajorItemsConfiguration
from randovania.layout.trick_level import LayoutTrickLevel, TrickLevelConfiguration
from randovania.resolver import debug
from randovania.resolver.state import State

_items_to_not_add_in_minimal_logic = {
    # Dark Visor
    10,

    # Dark Suit, Light Suit
    13,
    14,

    # Grapple, Screw Attack
    23,
    27,

    # Dark Agon Temple Keys
    32, 33, 34,

    # Dark Torvus Temple Keys
    35, 36, 37,

    # Sky Temple Keys
    29, 30, 31, 101, 102, 103, 104, 105, 106
}

_minimal_logic_custom_item_count = {
    # Energy Tank
    42: 14,
    # Power Bomb
    43: 10,
    # Missile
    44: 100,
    # Dark Ammo
    45: 100,
    # Light Ammo
    46: 100
}

_events_for_vanilla_item_loss_from_ship = {
    2,
    4,
    71,
    78,
}


def trick_resources_for_configuration(configuration: TrickLevelConfiguration,
                                      resource_database: ResourceDatabase,
                                      ) -> CurrentResources:
    """
    :param configuration:
    :param resource_database:
    :return:
    """

    static_resources = {}

    for trick in resource_database.trick:
        if configuration.minimal_logic:
            level = LayoutTrickLevel.HYPERMODE
        else:
            level = configuration.level_for_trick(trick)
        static_resources[trick] = level.as_number

    return static_resources


def _add_minimal_logic_initial_resources(resources: CurrentResources,
                                         resource_database: ResourceDatabase,
                                         major_items: MajorItemsConfiguration,
                                         ) -> None:
    # TODO: this function assumes we're talking about Echoes
    for event in resource_database.event:
        # Ignoring these events:
        # Dark Samus 3 and 4 (93), otherwise we're done automatically (TODO: get this from database)
        # Chykka (28), otherwise we can't collect Dark Visor
        if event.index not in {28, 93}:
            resources[event] = 1

    item_db = default_database.item_database_for_game(RandovaniaGame.PRIME2)

    items_to_skip = copy.copy(_items_to_not_add_in_minimal_logic)
    if major_items.items_state[item_db.major_items["Progressive Grapple"]].num_shuffled_pickups == 0:
        items_to_skip.remove(23)
    if major_items.items_state[item_db.major_items["Progressive Suit"]].num_shuffled_pickups == 0:
        items_to_skip.remove(13)

    for item in resource_database.item:
        if item.index not in items_to_skip:
            resources[item] = _minimal_logic_custom_item_count.get(item.index, 1)


def calculate_starting_state(game: GameDescription, patches: GamePatches) -> "State":
    starting_node = game.world_list.resolve_teleporter_connection(patches.starting_location)
    initial_resources = copy.copy(patches.starting_items)

    if isinstance(starting_node, PlayerShipNode):
        initial_resources[starting_node.resource()] = 1

    initial_game_state = game.initial_states.get("Default")
    if initial_game_state is not None:
        add_resource_gain_to_current_resources(initial_game_state, initial_resources)

    starting_state = State(
        initial_resources,
        (),
        99 + (100 * initial_resources.get(game.resource_database.energy_tank, 0)),
        starting_node,
        patches,
        None,
        game.resource_database,
        game.world_list,
    )

    # Being present with value 0 is troublesome since this dict is used for a simplify_requirements later on
    keys_to_remove = [resource for resource, quantity in initial_resources.items() if quantity == 0]
    for resource in keys_to_remove:
        del initial_resources[resource]

    return starting_state


def version_resources_for_game(resource_database: ResourceDatabase) -> CurrentResources:
    # All version differences are patched out from the game
    return {
        resource: 1 if resource.long_name == "NTSC" else 0
        for resource in resource_database.version
    }


def misc_resources_for_configuration(configuration: EchoesConfiguration,
                                     resource_database: ResourceDatabase) -> CurrentResources:
    enabled_resources = {
        # Allow Vanilla X
        19, 20, 21, 22, 23, 24, 25
    }
    if configuration.elevators == LayoutElevators.VANILLA:
        # Vanilla Great Temple Emerald Translator Gate
        enabled_resources.add(18)

    return {
        resource: 1 if resource.index in enabled_resources else 0
        for resource in resource_database.misc
    }


def logic_bootstrap(configuration: EchoesConfiguration,
                    game: GameDescription,
                    patches: GamePatches,
                    ) -> Tuple[GameDescription, State]:
    """
    Core code for starting a new Logic/State.
    :param configuration:
    :param game:
    :param patches:
    :return:
    """
    game = copy.deepcopy(game)
    starting_state = calculate_starting_state(game, patches)

    if configuration.trick_level.minimal_logic:
        _add_minimal_logic_initial_resources(starting_state.resources,
                                             game.resource_database,
                                             configuration.major_items_configuration)

    static_resources = trick_resources_for_configuration(configuration.trick_level,
                                                         game.resource_database)
    static_resources.update(version_resources_for_game(game.resource_database))
    static_resources.update(misc_resources_for_configuration(configuration, game.resource_database))

    for resource, quantity in static_resources.items():
        starting_state.resources[resource] = quantity

    game.patch_requirements(starting_state.resources, configuration.damage_strictness.value)

    return game, starting_state



from dataclasses import dataclass


@dataclass(frozen=True, order=True)
class AreaLocation:
    world_asset_id: int
    area_asset_id: int

    @property
    def as_json(self) -> dict:
        return {
            "world_asset_id": self.world_asset_id,
            "area_asset_id": self.area_asset_id,
        }

    @classmethod
    def from_json(cls, value: dict) -> "AreaLocation":
        return cls(value["world_asset_id"], value["area_asset_id"])

    def __repr__(self):
        return "world 0x{:02X}/area 0x{:02X}".format(self.world_asset_id, self.area_asset_id)



import copy
import dataclasses
from typing import Dict, Iterator, Tuple, List

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.item_database import ItemDatabase
from randovania.game_description.item.major_item import MajorItem
from randovania.layout.major_item_state import MajorItemState

RANDOM_STARTING_ITEMS_LIMIT = 31


@dataclasses.dataclass(frozen=True)
class MajorItemsConfiguration(BitPackValue):
    items_state: Dict[MajorItem, MajorItemState]
    default_items: Dict[ItemCategory, MajorItem]
    minimum_random_starting_items: int
    maximum_random_starting_items: int

    @property
    def as_json(self) -> dict:
        return {
            "items_state": {
                major_item.name: state.as_json
                for major_item, state in self.items_state.items()
            },
            "default_items": {
                category.value: item.name
                for category, item in self.default_items.items()
            },
            "minimum_random_starting_items": self.minimum_random_starting_items,
            "maximum_random_starting_items": self.maximum_random_starting_items,
        }

    @classmethod
    def from_json(cls, value: dict, item_database: ItemDatabase) -> "MajorItemsConfiguration":
        items_state = {}
        for name, item in item_database.major_items.items():
            if name in value["items_state"]:
                state = MajorItemState.from_json(value["items_state"][name])
            else:
                state = MajorItemState()
            items_state[item] = state

        default_items = {}
        for category, options in item_database.default_items.items():
            default_items[category] = item_database.major_items[value["default_items"][category.value]]
            if default_items[category] not in options:
                raise ValueError(f"Category {category} has {default_items[category]} as default item, "
                                 f"but that's not a valid option.")

        return cls(
            items_state=items_state,
            default_items=default_items,
            minimum_random_starting_items=value["minimum_random_starting_items"],
            maximum_random_starting_items=value["maximum_random_starting_items"],
        )

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        reference: MajorItemsConfiguration = metadata["reference"]

        result: List[Tuple[int, MajorItem, MajorItemState]] = []
        for i, (item, state) in enumerate(self.items_state.items()):
            if state != reference.items_state[item]:
                result.append((i, item, state))

        yield len(result), len(self.items_state)
        for index, _, _ in result:
            yield index, len(self.items_state)

        for index, item, state in result:
            yield from state.bit_pack_encode(item)

        # default_items
        for category, default in self.default_items.items():
            all_major = [major for major in reference.items_state.keys() if major.item_category == category]
            yield from bitpacking.pack_array_element(default, all_major)

        # random starting items
        yield self.minimum_random_starting_items, RANDOM_STARTING_ITEMS_LIMIT
        yield self.maximum_random_starting_items, RANDOM_STARTING_ITEMS_LIMIT

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "MajorItemsConfiguration":
        reference: MajorItemsConfiguration = metadata["reference"]

        num_items = decoder.decode_single(len(reference.items_state))
        indices_with_custom = {
            decoder.decode_single(len(reference.items_state))
            for _ in range(num_items)
        }

        items_state = {}

        for index, item in enumerate(reference.items_state.keys()):
            if index in indices_with_custom:
                items_state[item] = MajorItemState.bit_pack_unpack(decoder, item)
            else:
                items_state[item] = reference.items_state[item]

        # default_items
        default_items = {}
        for category in reference.default_items.keys():
            all_major = [major for major in reference.items_state.keys() if major.item_category == category]
            default_items[category] = decoder.decode_element(all_major)

        # random starting items
        minimum, maximum = decoder.decode(RANDOM_STARTING_ITEMS_LIMIT, RANDOM_STARTING_ITEMS_LIMIT)

        return cls(items_state,
                   default_items=default_items,
                   minimum_random_starting_items=minimum,
                   maximum_random_starting_items=maximum)

    def replace_state_for_item(self, item: MajorItem, state: MajorItemState) -> "MajorItemsConfiguration":
        return self.replace_states({
            item: state
        })

    def replace_states(self, new_states: Dict[MajorItem, MajorItemState]) -> "MajorItemsConfiguration":
        """
        Creates a copy of this MajorItemsConfiguration where the state of all given items are replaced by the given
        states.
        :param new_states:
        :return:
        """
        items_state = copy.copy(self.items_state)

        for item, state in new_states.items():
            items_state[item] = state

        return dataclasses.replace(self, items_state=items_state)

    def replace_default_item(self, category: ItemCategory, item: MajorItem) -> "MajorItemsConfiguration":
        """
        Creates a copy of this MajorItemsConfiguration where the default item for the given category
        is replaced by the given item.
        :param category:
        :param item:
        :return:
        """
        default_items = copy.copy(self.default_items)
        default_items[category] = item
        return dataclasses.replace(self, default_items=default_items)

    def calculate_provided_ammo(self) -> Dict[int, int]:
        result: Dict[int, int] = {}

        for item, state in self.items_state.items():
            total_pickups = state.num_shuffled_pickups + state.num_included_in_starting_items
            if state.include_copy_in_original_location:
                total_pickups += 1

            for ammo_index, ammo_count in zip(item.ammo_index, state.included_ammo):
                result[ammo_index] = result.get(ammo_index, 0) + ammo_count * total_pickups

        return result

    def dangerous_settings(self) -> List[str]:
        result = []
        for major_item, state in self.items_state.items():
            if (major_item.warning is not None and "EXPERIMENTAL" in major_item.warning
                    and state.num_included_in_starting_items == 0):
                result.append(f"Not starting with {major_item.name}")

        return result



import dataclasses

from PySide2 import QtCore

from randovania.gui.generated.preset_echoes_hints_ui import Ui_PresetEchoesHints
from randovania.gui.lib.common_qt_lib import set_combo_with_value
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.hint_configuration import SkyTempleKeyHintMode
from randovania.layout.preset import Preset


class PresetEchoesHints(PresetTab, Ui_PresetEchoesHints):
    _editor: PresetEditor

    def __init__(self, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        self._editor = editor

        self.hint_layout.setAlignment(QtCore.Qt.AlignTop)

        for i, stk_hint_mode in enumerate(SkyTempleKeyHintMode):
            self.hint_sky_temple_key_combo.setItemData(i, stk_hint_mode)

        self.hint_sky_temple_key_combo.currentIndexChanged.connect(self._on_stk_combo_changed)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def _on_stk_combo_changed(self, new_index: int):
        with self._editor as editor:
            editor.set_configuration_field(
                "hints",
                dataclasses.replace(editor.configuration.hints,
                                    sky_temple_keys=self.hint_sky_temple_key_combo.currentData()))

    def on_preset_changed(self, preset: Preset):
        set_combo_with_value(self.hint_sky_temple_key_combo, preset.configuration.hints.sky_temple_keys)



import datetime
import json
from unittest.mock import MagicMock, patch, ANY

import flask
import pytest

from randovania.network_common.error import InvalidSession
from randovania.server import user_session
from randovania.server.database import User


def test_setup_app():
    user_session.setup_app(MagicMock())


@patch("requests_oauthlib.OAuth2Session.fetch_token", autospec=True)
@pytest.mark.parametrize("existing", [False, True])
def test_login_with_discord(mock_fetch_token: MagicMock, clean_database, flask_app, existing):
    # Setup
    sio = MagicMock()
    session = {}
    sio.session.return_value.__enter__.return_value = session
    sio.get_session.return_value = session
    mock_fetch_token.return_value = "access_token"
    sio.fernet_encrypt.encrypt.return_value = b"encrypted"

    discord_user = sio.discord.fetch_user.return_value
    discord_user.id = 1234
    discord_user.name = "A Name"

    if existing:
        User.create(discord_id=discord_user.id, name="Someone else")

    # Run
    with flask_app.test_request_context():
        result = user_session.login_with_discord(sio, "code")

    # Assert
    mock_fetch_token.assert_called_once_with(
        ANY,
        "https://discord.com/api/oauth2/token",
        code="code",
        client_secret=sio.app.config["DISCORD_CLIENT_SECRET"],
    )
    user = User.get(User.discord_id == 1234)
    assert user.name == "A Name"

    assert session == {
        "user-id": user.id,
        "discord-access-token": "access_token",
    }
    assert result == {
        "user": user.as_json,
        "sessions": [],
        "encoded_session_b85": b'Wo~0~d2n=PWB',
    }


@patch("randovania.server.user_session._create_client_side_session", autospec=True)
def test_restore_user_session_with_discord(mock_create_session: MagicMock,
                                           flask_app, fernet, clean_database):
    sio = MagicMock()
    sio.fernet_encrypt = fernet

    user: User = User.create(
        id=1234,
        discord_id=5678,
        name="The Name"
    )

    session = {
        "user-id": 1234,
        "discord-access-token": "access-token",
    }
    enc_session = fernet.encrypt(json.dumps(session).encode("utf-8"))

    # Run
    with flask_app.test_request_context():
        flask.request.sid = 7890
        result = user_session.restore_user_session(sio, enc_session, None)

    # Assert
    sio.get_server.return_value.save_session.assert_called_once_with(7890, session)
    mock_create_session.assert_called_once_with(sio, user)
    assert result is mock_create_session.return_value


def test_login_with_guest(flask_app, clean_database, mocker):
    # Setup
    mocker.patch("randovania.server.user_session._get_now", return_value=datetime.datetime(year=2020, month=9, day=4))
    mock_create_session = mocker.patch("randovania.server.user_session._create_client_side_session", autospec=True)
    enc_request = b"encrypted stuff"

    sio = MagicMock()
    sio.guest_encrypt.decrypt.return_value = json.dumps({
        "name": "Someone",
        "date": '2020-09-05T17:12:09.941661',
    }).encode("utf-8")

    with flask_app.test_request_context():
        flask.request.sid = 7890
        result = user_session.login_with_guest(sio, enc_request)

    # Assert
    sio.guest_encrypt.decrypt.assert_called_once_with(enc_request)
    user: User = User.get_by_id(1)
    assert user.name == "Guest: Someone"

    mock_create_session.assert_called_once_with(sio, user)
    assert result is mock_create_session.return_value


def test_logout(flask_app, mocker):
    mock_emit_user_session_update: MagicMock = mocker.patch("randovania.server.user_session._emit_user_session_update",
                                                            autospec=True)
    session = {
        "user-id": 1234,
        "discord-access-token": "access_token",
    }
    sio = MagicMock()
    sio.session.return_value.__enter__.return_value = session

    # Run
    with flask_app.test_request_context():
        user_session.logout(sio)

    # Assert
    assert session == {}
    sio.leave_game_session.assert_called_once_with()
    mock_emit_user_session_update.assert_not_called()


def test_restore_user_session_invalid_key(flask_app, fernet):
    sio = MagicMock()
    sio.fernet_encrypt = fernet

    with pytest.raises(InvalidSession):
        user_session.restore_user_session(sio, b"", None)
        pass



import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.games.game import RandovaniaGame
from randovania.layout.trick_level import TrickLevelConfiguration, LayoutTrickLevel


@pytest.fixture(
    params=[
        {"encoded": b'\x00\x00', "json": {"minimal_logic": False, "specific_levels": {}}},
        {"encoded": b'\x80', "json": {"minimal_logic": True, "specific_levels": {}}},
        {"encoded": b'X\x00\x00', "json": {"minimal_logic": False, "specific_levels": {"Dash": "expert"}}},
        {"encoded": b'f3\x00\x00', "json": {"minimal_logic": False, "specific_levels": {
            i: "hypermode"
            for i in ["Dash", "BombJump", "Movement", "BSJ"]
        }}},
    ],
    name="configuration_with_data")
def _configuration_with_data(request, mocker, echoes_game_description):
    tricks = echoes_game_description.resource_database.trick[:14]
    mocker.patch("randovania.layout.trick_level._all_tricks", return_value=tricks)
    return request.param["encoded"], TrickLevelConfiguration.from_json(request.param["json"],
                                                                       game=RandovaniaGame.PRIME2)


def test_decode(configuration_with_data):
    # Setup
    data, expected = configuration_with_data

    # Run
    decoder = BitPackDecoder(data)
    result = TrickLevelConfiguration.bit_pack_unpack(decoder, {
        "reference": TrickLevelConfiguration(False, {}, RandovaniaGame.PRIME2),
    })

    # Assert
    assert result == expected


def test_encode(configuration_with_data):
    expected, value = configuration_with_data

    # Run
    result = bitpacking._pack_encode_results([
        (value_argument, value_format)
        for value_argument, value_format in value.bit_pack_encode({})
    ])

    # Assert
    assert result == expected


def test_encode_no_tricks_are_removed():
    from_json = TrickLevelConfiguration.from_json({"minimal_logic": False, "specific_levels": {"Dash": "disabled"}},
                                                  game=RandovaniaGame.PRIME2)

    encoded = bitpacking._pack_encode_results([
        (value_argument, value_format)
        for value_argument, value_format in from_json.bit_pack_encode({})
    ])

    assert encoded == b'\x00\x00\x00\x00'

    decoder = BitPackDecoder(encoded)
    decoded = TrickLevelConfiguration.bit_pack_unpack(
        decoder, {"reference": TrickLevelConfiguration(False, {}, RandovaniaGame.PRIME2), })

    assert decoded.specific_levels == {}


def test_set_level_for_trick_remove(echoes_resource_database):
    trick = echoes_resource_database.trick[0]
    config = TrickLevelConfiguration(False, {}, RandovaniaGame.PRIME2)

    assert config.level_for_trick(trick) == LayoutTrickLevel.DISABLED

    config = config.set_level_for_trick(trick, LayoutTrickLevel.ADVANCED)
    assert config.level_for_trick(trick) == LayoutTrickLevel.ADVANCED

    config = config.set_level_for_trick(trick, LayoutTrickLevel.DISABLED)
    assert config.level_for_trick(trick) == LayoutTrickLevel.DISABLED


def test_pretty_description_minimal_logic():
    config = TrickLevelConfiguration(True, {}, RandovaniaGame.PRIME2)
    assert config.pretty_description == "Minimal Logic"


@pytest.mark.parametrize(["levels", "expected"], [
    ({}, "All at Disabled"),
    ({i: LayoutTrickLevel.HYPERMODE for i in ["Dash", "BombJump", "Movement", "BSJ"]},
     "20 at Disabled, 4 at Hypermode"),
    ({"Dash": LayoutTrickLevel.HYPERMODE, "BombJump": LayoutTrickLevel.HYPERMODE,
      "AirUnderwater": LayoutTrickLevel.ADVANCED},
     "21 at Disabled, 1 at Advanced, 2 at Hypermode"),
])
def test_pretty_description_tricks_echoes(levels, expected):
    config = TrickLevelConfiguration(False, levels, RandovaniaGame.PRIME2)
    assert config.pretty_description == expected




from unittest.mock import MagicMock

import pytest

from randovania.game_description import game_description
from randovania.game_description.area import Area
from randovania.game_description.node import Node
from randovania.game_description.requirements import Requirement


@pytest.mark.parametrize(["danger_a", "danger_b", "expected_result"], [
    ([], [], []),
    (["a"], [], ["a"]),
    ([], ["b"], ["b"]),
    (["a"], ["b"], ["a", "b"]),
    (["a"], ["a"], ["a"]),
])
def test_calculate_dangerous_resources(danger_a, danger_b, expected_result):
    set_a: Requirement = MagicMock()
    set_b: Requirement = MagicMock()

    set_a.as_set.dangerous_resources = danger_a
    set_b.as_set.dangerous_resources = danger_b

    n1: Node = "n1"
    n2: Node = "n2"

    area_a = Area(
        "area_a", False, 0, 0, True, [n1, n2],
        {
            n1: {
                n2: set_a
            },
            n2: {}
        }
    )
    area_b = Area(
        "area_b", True, 0, 0, True, [n1, n2],
        {
            n1: {},
            n2: {
                n1: set_b
            }
        }
    )

    # Run
    result = game_description._calculate_dangerous_resources_in_areas([area_a, area_b])

    # Assert
    assert set(result) == set(expected_result)



import pytest

from randovania.game_description.node import LogbookNode, LoreType
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_info import convert_resource_gain_to_current_resources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo


@pytest.fixture(
    params=[False, True],
    name="logbook_node")
def _logbook_node(request):
    has_translator = request.param
    scan_visor = ItemResourceInfo(1, "Scan", "S", 1, None)
    translator = ItemResourceInfo(2, "Translator", "T", 1, None)

    node = LogbookNode("Logbook", False, None, 0,
                       1000, scan_visor, LoreType.LUMINOTH_LORE,
                       translator if has_translator else None, None)

    return has_translator, scan_visor, translator, node


def test_logbook_node_requirements_to_leave(logbook_node,
                                            empty_patches):
    # Setup
    has_translator, scan_visor, translator, node = logbook_node

    # Run
    to_leave = node.requirement_to_leave(empty_patches, {})

    # Assert
    assert not to_leave.satisfied({}, 99)
    assert to_leave.satisfied({scan_visor: 1}, 99) != has_translator
    assert to_leave.satisfied({scan_visor: 1, translator: 1}, 99)


def test_logbook_node_can_collect(logbook_node,
                                  empty_patches):
    # Setup
    has_translator, scan_visor, translator, node = logbook_node

    assert not node.can_collect(empty_patches, {}, ())
    assert node.can_collect(empty_patches, {scan_visor: 1}, ()) != has_translator
    assert node.can_collect(empty_patches, {scan_visor: 1, translator: 1}, ())

    resource = node.resource()
    assert not node.can_collect(empty_patches, {resource: 1}, ())
    assert not node.can_collect(empty_patches, {resource: 1, scan_visor: 1}, ())
    assert not node.can_collect(empty_patches, {resource: 1, scan_visor: 1, translator: 1}, ())


def test_logbook_node_resource_gain_on_collect(logbook_node,
                                               empty_patches):
    # Setup
    node = logbook_node[-1]

    # Run
    gain = node.resource_gain_on_collect(empty_patches, {}, ())

    # Assert
    assert convert_resource_gain_to_current_resources(gain) == {node.resource(): 1}



import dataclasses
import json
from typing import Optional, List, Iterator, Tuple, Union

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder, BitPackValue
from randovania.games.game import RandovaniaGame
from randovania.layout.corruption_configuration import CorruptionConfiguration
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.prime_configuration import PrimeConfiguration


def _dictionary_byte_hash(data: dict) -> int:
    return bitpacking.single_byte_hash(json.dumps(data, separators=(',', ':')).encode("UTF-8"))


_game_to_config = {
    RandovaniaGame.PRIME1: PrimeConfiguration,
    RandovaniaGame.PRIME2: EchoesConfiguration,
    RandovaniaGame.PRIME3: CorruptionConfiguration,
}


@dataclasses.dataclass(frozen=True)
class Preset(BitPackValue):
    name: str
    description: str
    base_preset_name: Optional[str]
    game: RandovaniaGame
    configuration: Union[EchoesConfiguration, CorruptionConfiguration]

    @property
    def as_json(self) -> dict:
        return {
            "name": self.name,
            "description": self.description,
            "base_preset_name": self.base_preset_name,
            "game": self.game.value,
            "configuration": self.configuration.as_json,
        }

    @classmethod
    def from_json_dict(cls, value) -> "Preset":
        game = RandovaniaGame(value["game"])
        return Preset(
            name=value["name"],
            description=value["description"],
            base_preset_name=value["base_preset_name"],
            game=game,
            configuration=_game_to_config[game].from_json(value["configuration"]),
        )

    def dangerous_settings(self) -> List[str]:
        return self.configuration.dangerous_settings()

    def is_same_configuration(self, other: "Preset") -> bool:
        return self.configuration == other.configuration

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        from randovania.interface_common.preset_manager import PresetManager
        manager: PresetManager = metadata["manager"]

        # Is this a custom preset?
        is_custom_preset = self.base_preset_name is not None
        if is_custom_preset:
            reference_versioned = manager.included_preset_with_name(self.base_preset_name)
            if reference_versioned is None:
                reference_versioned = manager.default_preset_for_game(self.game)
            reference = reference_versioned.get_preset()
        else:
            reference = self

        included_presets = [versioned.get_preset() for versioned in manager.included_presets]

        yield from bitpacking.encode_bool(is_custom_preset)
        yield from bitpacking.pack_array_element(reference, included_presets)
        if is_custom_preset:
            yield from self.configuration.bit_pack_encode({"reference": reference.configuration})
        yield _dictionary_byte_hash(self.configuration.game_data), 256

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "Preset":
        from randovania.interface_common.preset_manager import PresetManager
        manager: PresetManager = metadata["manager"]

        included_presets = [versioned.get_preset() for versioned in manager.included_presets]

        is_custom_preset = bitpacking.decode_bool(decoder)
        reference = decoder.decode_element(included_presets)
        if is_custom_preset:
            preset = Preset(
                name="{} Custom".format(reference.name),
                description="A customized preset.",
                base_preset_name=reference.name,
                game=reference.game,
                configuration=reference.configuration.bit_pack_unpack(decoder, {"reference": reference.configuration}),
            )
        else:
            preset = reference

        included_data_hash = decoder.decode_single(256)
        expected_data_hash = _dictionary_byte_hash(preset.configuration.game_data)
        if included_data_hash != expected_data_hash:
            raise ValueError("Given permalink is for a Randovania database with hash '{}', "
                             "but current database has hash '{}'.".format(included_data_hash, expected_data_hash))
        return preset



import copy
import re
from typing import List, Dict, Iterator, Tuple, Iterable

from randovania.game_description.area import Area
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.dock import DockConnection
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import Node, DockNode, TeleporterNode, PickupNode, PlayerShipNode
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.resource_info import CurrentResources
from randovania.game_description.world import World


class WorldList:
    worlds: List[World]

    _nodes_to_area: Dict[Node, Area]
    _nodes_to_world: Dict[Node, World]
    _ids_to_area: Dict[AreaLocation, Area]
    _nodes: Tuple[Node, ...]

    def __deepcopy__(self, memodict):
        return WorldList(
            worlds=copy.deepcopy(self.worlds, memodict),
        )

    def __init__(self, worlds: List[World]):
        self.worlds = worlds
        self.refresh_node_cache()

    def refresh_node_cache(self):
        self._nodes_to_area, self._nodes_to_world, self._ids_to_area = _calculate_nodes_to_area_world(self.worlds)
        self._nodes = tuple(self._iterate_over_nodes())

    def _iterate_over_nodes(self) -> Iterator[Node]:
        for world in self.worlds:
            yield from world.all_nodes

    def world_with_name(self, world_name: str) -> World:
        for world in self.worlds:
            if world.name == world_name or world.dark_name == world_name:
                return world
        raise KeyError("Unknown name: {}".format(world_name))

    def world_by_asset_id(self, asset_id: int) -> World:
        for world in self.worlds:
            if world.world_asset_id == asset_id:
                return world
        raise KeyError("Unknown asset_id: {}".format(asset_id))

    def world_with_area(self, area: Area) -> World:
        for world in self.worlds:
            if area in world.areas:
                return world
        raise KeyError("Unknown area: {}".format(area))

    @property
    def all_areas(self) -> Iterator[Area]:
        for world in self.worlds:
            yield from world.areas

    @property
    def all_nodes(self) -> Tuple[Node, ...]:
        return self._nodes

    @property
    def num_pickup_nodes(self) -> int:
        return sum(1 for node in self.all_nodes if isinstance(node, PickupNode))

    @property
    def all_worlds_areas_nodes(self) -> Iterable[Tuple[World, Area, Node]]:
        for world in self.worlds:
            for area in world.areas:
                for node in area.nodes:
                    yield world, area, node

    def world_name_from_area(self, area: Area, distinguish_dark_aether: bool = False) -> str:
        world = self.world_with_area(area)

        if distinguish_dark_aether:
            return world.correct_name(area.in_dark_aether)
        else:
            return world.name

    def world_name_from_node(self, node: Node, distinguish_dark_aether: bool = False) -> str:
        return self.world_name_from_area(self.nodes_to_area(node), distinguish_dark_aether)

    def area_name(self, area: Area, separator: str = " - ", distinguish_dark_aether: bool = True) -> str:
        return "{}{}{}".format(
            self.world_name_from_area(area, distinguish_dark_aether),
            separator,
            area.name)

    def node_name(self, node: Node, with_world=False, distinguish_dark_aether: bool = False) -> str:
        prefix = "{}/".format(self.world_name_from_node(node, distinguish_dark_aether)) if with_world else ""
        return "{}{}/{}".format(prefix, self.nodes_to_area(node).name, node.name)

    def node_from_name(self, name: str) -> Node:
        match = re.match("(?:([^/]+)/)?([^/]+)/([^/]+)", name)
        if match is None:
            raise ValueError("Invalid name: {}".format(name))

        world_name, area_name, node_name = match.group(1, 2, 3)
        for world in self.worlds:
            if world_name is not None and world_name not in (world.name, world.dark_name):
                continue

            for area in world.areas:
                if area.name != area_name:
                    continue

                for node in area.nodes:
                    if node.name == node_name:
                        return node

        raise ValueError("Unknown name: {}".format(name))

    def nodes_to_world(self, node: Node) -> World:
        return self._nodes_to_world[node]

    def nodes_to_area(self, node: Node) -> Area:
        return self._nodes_to_area[node]

    def resolve_dock_connection(self, world: World, connection: DockConnection) -> Node:
        target_area = world.area_by_asset_id(connection.area_asset_id)
        return target_area.node_with_dock_index(connection.dock_index)

    def resolve_dock_node(self, node: DockNode, patches: GamePatches) -> Node:
        world = self.nodes_to_world(node)
        original_area = self.nodes_to_area(node)

        connection = patches.dock_connection.get((original_area.area_asset_id, node.dock_index),
                                                 node.default_connection)
        return self.resolve_dock_connection(world, connection)

    def resolve_teleporter_node(self, node: TeleporterNode, patches: GamePatches) -> Node:
        connection = patches.elevator_connection.get(node.teleporter_instance_id, node.default_connection)
        return self.resolve_teleporter_connection(connection)

    def resolve_teleporter_connection(self, connection: AreaLocation) -> Node:
        area = self.area_by_area_location(connection)
        if area.default_node_index is None:
            raise IndexError("Area '{}' does not have a default_node_index".format(area.name))
        try:
            return area.nodes[area.default_node_index]
        except IndexError:
            raise IndexError("Area '{}' default_node_index ({}), but there's only {} nodes".format(
                area.name, area.default_node_index, len(area.nodes)))

    def connections_from(self, node: Node, patches: GamePatches) -> Iterator[Tuple[Node, Requirement]]:
        """
        Queries all nodes from other areas you can go from a given node. Aka, doors and teleporters
        :param patches:
        :param node:
        :return: Generator of pairs Node + Requirement for going to that node
        """
        if isinstance(node, DockNode):
            # TODO: respect is_blast_shield: if already opened once, no requirement needed.
            # Includes opening form behind with different criteria
            try:
                target_node = self.resolve_dock_node(node, patches)
                original_area = self.nodes_to_area(node)
                dock_weakness = patches.dock_weakness.get((original_area.area_asset_id, node.dock_index),
                                                          node.default_dock_weakness)

                yield target_node, dock_weakness.requirement
            except IndexError:
                # TODO: fix data to not having docks pointing to nothing
                yield None, Requirement.impossible()

        if isinstance(node, TeleporterNode):
            try:
                yield self.resolve_teleporter_node(node, patches), Requirement.trivial()
            except IndexError:
                # TODO: fix data to not have teleporters pointing to areas with invalid default_node_index
                print("Teleporter is broken!", node)
                yield None, Requirement.impossible()

        if isinstance(node, PlayerShipNode):
            for other_node in self.all_nodes:
                if isinstance(other_node, PlayerShipNode) and other_node != node:
                    yield other_node, other_node.is_unlocked

    def area_connections_from(self, node: Node) -> Iterator[Tuple[Node, Requirement]]:
        """
        Queries all nodes from the same area you can go from a given node.
        :param node:
        :return: Generator of pairs Node + Requirement for going to that node
        """
        area = self.nodes_to_area(node)
        for target_node, requirements in area.connections[node].items():
            yield target_node, requirements

    def potential_nodes_from(self, node: Node, patches: GamePatches) -> Iterator[Tuple[Node, Requirement]]:
        """
        Queries all nodes you can go from a given node, checking doors, teleporters and other nodes in the same area.
        :param node:
        :param patches:
        :return: Generator of pairs Node + Requirement for going to that node
        """
        yield from self.connections_from(node, patches)
        yield from self.area_connections_from(node)

    def patch_requirements(self, static_resources: CurrentResources, damage_multiplier: float) -> None:
        """
        Patches all Node connections, assuming the given resources will never change their quantity.
        This is removes all checking for tricks and difficulties in runtime since these never change.
        All damage requirements are multiplied by the given multiplier.
        :param static_resources:
        :param damage_multiplier:
        :return:
        """
        for world in self.worlds:
            for area in world.areas:
                for node in area.nodes:
                    if isinstance(node, DockNode):
                        requirement = node.default_dock_weakness.requirement
                        object.__setattr__(node.default_dock_weakness, "requirement",
                                           requirement.patch_requirements(static_resources,
                                                                          damage_multiplier).simplify())
                for connections in area.connections.values():
                    for target, value in connections.items():
                        connections[target] = value.patch_requirements(static_resources, damage_multiplier).simplify()

    def area_by_area_location(self, location: AreaLocation) -> Area:
        return self._ids_to_area[location]

    def world_by_area_location(self, location: AreaLocation) -> World:
        return self.world_by_asset_id(location.world_asset_id)

    def area_to_area_location(self, area: Area) -> AreaLocation:
        for world in self.worlds:
            result = AreaLocation(world.world_asset_id, area.area_asset_id)
            if result in self._ids_to_area:
                return result
        raise RuntimeError(f"Unknown area: {area}")

    def node_to_area_location(self, node: Node) -> AreaLocation:
        return AreaLocation(
            world_asset_id=self.nodes_to_world(node).world_asset_id,
            area_asset_id=self.nodes_to_area(node).area_asset_id,
        )

    def add_new_node(self, area: Area, node: Node):
        self._nodes_to_area[node] = area
        self._nodes_to_world[node] = self.world_with_area(area)


def _calculate_nodes_to_area_world(worlds: Iterable[World]):
    nodes_to_area = {}
    nodes_to_world = {}
    ids_to_area = {}

    for world in worlds:
        for area in world.areas:
            ids_to_area[AreaLocation(world.world_asset_id, area.area_asset_id)] = area
            for node in area.nodes:
                if node in nodes_to_area:
                    raise ValueError(
                        "Trying to map {} to {}, but already mapped to {}".format(
                            node, area, nodes_to_area[node]))
                nodes_to_area[node] = area
                nodes_to_world[node] = world

    return nodes_to_area, nodes_to_world, ids_to_area



import copy
import dataclasses
from random import Random
from typing import List, Tuple, Callable, TypeVar, Set, Dict, FrozenSet, Union, Iterator, Optional

from randovania.game_description import node_search
from randovania.game_description.area import Area
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint, HintType, PrecisionPair, HintLocationPrecision, HintItemPrecision, \
    HintRelativeAreaName, RelativeDataArea, RelativeDataItem
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import LogbookNode, PickupNode
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.generator.filler.filler_library import should_have_hint, UnableToGenerate
from randovania.generator.filler.retcon import retcon_playthrough_filler, FillerConfiguration, PlayerState
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.resolver import bootstrap, debug, random_lib

T = TypeVar("T")


def _split_expansions(item_pool: List[PickupEntry]) -> Tuple[List[PickupEntry], List[PickupEntry]]:
    """

    :param item_pool:
    :return:
    """
    major_items = []
    expansions = []

    for pickup in item_pool:
        if pickup.item_category == ItemCategory.EXPANSION:
            expansions.append(pickup)
        else:
            major_items.append(pickup)

    return major_items, expansions


def _create_weighted_list(rng: Random,
                          current: List[T],
                          factory: Callable[[], List[T]],
                          ) -> List[T]:
    """
    Ensures we always have a non-empty list.
    :param rng:
    :param current:
    :param factory:
    :return:
    """
    if not current:
        current = factory()
        rng.shuffle(current)

    return current


def precision_pair_weighted_list() -> List[PrecisionPair]:
    tiers = {
        (HintLocationPrecision.DETAILED, HintItemPrecision.DETAILED): 5,
        (HintLocationPrecision.DETAILED, HintItemPrecision.PRECISE_CATEGORY): 2,
        (HintLocationPrecision.DETAILED, HintItemPrecision.GENERAL_CATEGORY): 1,

        (HintLocationPrecision.WORLD_ONLY, HintItemPrecision.DETAILED): 2,
        (HintLocationPrecision.WORLD_ONLY, HintItemPrecision.PRECISE_CATEGORY): 1,
    }

    hints = []
    for params, quantity in tiers.items():
        hints.extend([PrecisionPair(*params, include_owner=False)] * quantity)

    return hints


def _not_empty(it: Iterator) -> bool:
    return sum(1 for _ in it) > 0


def add_relative_hint(world_list: WorldList,
                      patches: GamePatches,
                      rng: Random,
                      target: PickupIndex,
                      target_precision: HintItemPrecision,
                      relative_type: HintLocationPrecision,
                      precise_distance: bool,
                      precision: Union[HintItemPrecision, HintRelativeAreaName],
                      max_distance: int,
                      ) -> Optional[Hint]:
    """
    Creates a relative hint.
    :return: Might be None, if no hint could be created.
    """
    target_node = node_search.pickup_index_to_node(world_list, target)
    target_area = world_list.nodes_to_area(target_node)
    distances = node_search.distances_to_node(world_list, target_node, patches=patches, cutoff=max_distance)

    def _major_pickups(area: Area) -> Iterator[PickupIndex]:
        for index in area.pickup_indices:
            t = patches.pickup_assignment.get(index)
            # FIXME: None should be ok, but this must be called after junk has been filled
            if t is not None:
                cat = t.pickup.item_category
                if cat.is_major_category or (cat != ItemCategory.EXPANSION
                                             and target_precision == HintItemPrecision.DETAILED):
                    yield index

    area_choices = {
        area: 1 / max(distance, 2)
        for area, distance in distances.items()
        if (distance > 0 and area.in_dark_aether == target_area.in_dark_aether
            and (relative_type == HintLocationPrecision.RELATIVE_TO_AREA or _not_empty(_major_pickups(area))))
    }
    if not area_choices:
        return None
    area = random_lib.select_element_with_weight(dict(sorted(area_choices.items(),
                                                             key=lambda a: a[0].area_asset_id)), rng)

    if relative_type == HintLocationPrecision.RELATIVE_TO_AREA:
        relative = RelativeDataArea(precise_distance, world_list.area_to_area_location(area),
                                    precision)
    elif relative_type == HintLocationPrecision.RELATIVE_TO_INDEX:
        relative = RelativeDataItem(precise_distance, rng.choice(list(_major_pickups(area))), precision)
    else:
        raise ValueError(f"Invalid relative_type: {relative_type}")

    precision_pair = PrecisionPair(relative_type, target_precision, include_owner=False, relative=relative)
    return Hint(HintType.LOCATION, precision_pair, target)


def _relative(relative_type: HintLocationPrecision,
              precise_distance: bool,
              precision: Union[HintItemPrecision, HintRelativeAreaName],
              max_distance: int,
              ) -> Callable[[PlayerState, GamePatches, Random, PickupIndex], Optional[Hint]]:
    def _wrapper(player_state: PlayerState, patches: GamePatches, rng: Random, target: PickupIndex):
        return add_relative_hint(player_state.game.world_list, patches, rng, target, HintItemPrecision.DETAILED,
                                 relative_type, precise_distance, precision, max_distance)

    return _wrapper


def _get_relative_hint_providers():
    return [
        _relative(HintLocationPrecision.RELATIVE_TO_AREA, True, HintRelativeAreaName.NAME, 4),
        _relative(HintLocationPrecision.RELATIVE_TO_INDEX, True, HintItemPrecision.DETAILED, 4),
        _relative(HintLocationPrecision.RELATIVE_TO_INDEX, True, HintItemPrecision.PRECISE_CATEGORY, 3),
    ]


def add_hints_precision(player_state: PlayerState,
                        patches: GamePatches,
                        rng: Random,
                        ) -> GamePatches:
    """
    Adds precision to all hints that are missing one.
    :param player_state:
    :param patches:
    :param rng:
    :return:
    """

    hints_to_replace: Dict[LogbookAsset, Hint] = {
        asset: hint
        for asset, hint in patches.hints.items()
        if hint.precision is None and hint.hint_type == HintType.LOCATION
    }

    relative_hint_providers = _get_relative_hint_providers()
    asset_ids = list(hints_to_replace.keys())
    rng.shuffle(asset_ids)

    while asset_ids and relative_hint_providers:
        new_hint = relative_hint_providers.pop()(player_state, patches, rng, hints_to_replace[asset_ids[-1]].target)
        if new_hint is not None:
            hints_to_replace[asset_ids.pop()] = new_hint

    # Add random precisions
    precisions = []
    for asset_id in asset_ids:
        precisions = _create_weighted_list(rng, precisions, precision_pair_weighted_list)
        precision = precisions.pop()

        hints_to_replace[asset_id] = dataclasses.replace(hints_to_replace[asset_id], precision=precision)

    # Replace the hints the in the patches
    return dataclasses.replace(patches, hints={
        asset: hints_to_replace.get(asset, hint)
        for asset, hint in patches.hints.items()
    })


def replace_hints_without_precision_with_jokes(patches: GamePatches,
                                               ) -> GamePatches:
    """
    Adds WRONG_GAME precision to all hints that are missing one precision.
    :param patches:
    :return:
    """

    hints_to_replace = {
        asset: Hint(HintType.JOKE, None)
        for asset, hint in patches.hints.items()
        if hint.precision is None and hint.hint_type == HintType.LOCATION
    }

    return dataclasses.replace(patches, hints={
        asset: hints_to_replace.get(asset, hint)
        for asset, hint in patches.hints.items()
    })


def fill_unassigned_hints(patches: GamePatches,
                          world_list: WorldList,
                          rng: Random,
                          scan_asset_initial_pickups: Dict[LogbookAsset, FrozenSet[PickupIndex]],
                          ) -> GamePatches:
    new_hints = copy.copy(patches.hints)

    # Get all LogbookAssets from the WorldList
    potential_hint_locations: Set[LogbookAsset] = {
        node.resource()
        for node in world_list.all_nodes
        if isinstance(node, LogbookNode)
    }
    for logbook in potential_hint_locations:
        if logbook not in scan_asset_initial_pickups:
            scan_asset_initial_pickups[logbook] = frozenset()

    # But remove these that already have hints
    potential_hint_locations -= patches.hints.keys()

    # Get interesting items to place hints for
    possible_indices = set(patches.pickup_assignment.keys())
    possible_indices -= {hint.target for hint in patches.hints.values() if hint.target is not None}
    possible_indices -= {index for index in possible_indices
                         if not should_have_hint(patches.pickup_assignment[index].pickup.item_category)}

    debug.debug_print("fill_unassigned_hints had {} decent indices for {} hint locations".format(
        len(possible_indices), len(potential_hint_locations)))

    if debug.debug_level() > 1:
        print(f"> Num pickups per asset:")
        for asset, pickups in scan_asset_initial_pickups.items():
            print(f"* {asset}: {len(pickups)} pickups")
        print("> Done.")

    # But if we don't have enough hints, just pick randomly from everything
    if len(possible_indices) < len(potential_hint_locations):
        possible_indices = {node.pickup_index
                            for node in world_list.all_nodes
                            if isinstance(node, PickupNode)}

    # Get an stable order
    ordered_possible_indices = list(sorted(possible_indices))
    ordered_potential_hint_locations = list(sorted(potential_hint_locations))

    num_logbooks: Dict[PickupIndex, int] = {
        index: sum(1 for indices in scan_asset_initial_pickups.values() if index in indices)
        for index in ordered_possible_indices
    }
    max_seen = max(num_logbooks.values())
    pickup_indices_weight: Dict[PickupIndex, int] = {
        index: max_seen - num_logbook
        for index, num_logbook in num_logbooks.items()
    }
    # Ensure all indices are present with at least weight 0
    for index in ordered_possible_indices:
        if index not in pickup_indices_weight:
            pickup_indices_weight[index] = 0

    for logbook in sorted(ordered_potential_hint_locations,
                          key=lambda r: len(scan_asset_initial_pickups[r]),
                          reverse=True):
        try:
            new_index = random_lib.select_element_with_weight(pickup_indices_weight, rng)
        except StopIteration:
            # If everything has weight 0, then just choose randomly.
            new_index = random_lib.random_key(pickup_indices_weight, rng)
            
        del pickup_indices_weight[new_index]

        new_hints[logbook] = Hint(HintType.LOCATION, None, new_index)
        debug.debug_print(f"Added hint at {logbook} for item at {new_index}")

    return dataclasses.replace(patches, hints=new_hints)


@dataclasses.dataclass(frozen=True)
class PlayerPool:
    game: GameDescription
    configuration: EchoesConfiguration
    patches: GamePatches
    pickups: List[PickupEntry]


@dataclasses.dataclass(frozen=True)
class FillerPlayerResult:
    game: GameDescription
    patches: GamePatches
    unassigned_pickups: List[PickupEntry]


@dataclasses.dataclass(frozen=True)
class FillerResults:
    player_results: Dict[int, FillerPlayerResult]
    action_log: Tuple[str, ...]


def run_filler(rng: Random,
               player_pools: Dict[int, PlayerPool],
               status_update: Callable[[str], None],
               ) -> FillerResults:
    """
    Runs the filler logic for the given configuration and item pool.
    Returns a GamePatches with progression items and hints assigned, along with all items in the pool
    that weren't assigned.

    :param player_pools:
    :param rng:
    :param status_update:
    :return:
    """

    player_states = []
    player_expansions: Dict[int, List[PickupEntry]] = {}

    for index, pool in player_pools.items():
        status_update(f"Creating state for player {index + 1}")
        major_items, player_expansions[index] = _split_expansions(pool.pickups)
        rng.shuffle(major_items)
        rng.shuffle(player_expansions[index])

        new_game, state = bootstrap.logic_bootstrap(pool.configuration, pool.game, pool.patches)
        new_game.patch_requirements(state.resources, pool.configuration.damage_strictness.value)

        major_configuration = pool.configuration.major_items_configuration
        player_states.append(PlayerState(
            index=index,
            game=new_game,
            initial_state=state,
            pickups_left=major_items,
            configuration=FillerConfiguration(
                randomization_mode=pool.configuration.available_locations.randomization_mode,
                minimum_random_starting_items=major_configuration.minimum_random_starting_items,
                maximum_random_starting_items=major_configuration.maximum_random_starting_items,
                indices_to_exclude=pool.configuration.available_locations.excluded_indices,
            ),
        ))

    try:
        filler_result, actions_log = retcon_playthrough_filler(rng, player_states, status_update=status_update)
    except UnableToGenerate as e:
        message = "{}\n\n{}".format(
            str(e),
            "\n\n".join(
                "#### Player {}\n{}".format(player.index + 1, player.current_state_report())
                for player in player_states
            ),
        )
        debug.debug_print(message)
        raise UnableToGenerate(message) from e

    results = {}

    for player_state, patches in filler_result.items():
        game = player_state.game

        if game.game == RandovaniaGame.PRIME2:
            # Since we haven't added expansions yet, these hints will always be for items added by the filler.
            full_hints_patches = fill_unassigned_hints(patches, game.world_list, rng,
                                                       player_state.scan_asset_initial_pickups)

            if player_pools[player_state.index].configuration.hints.item_hints:
                result = add_hints_precision(player_state, full_hints_patches, rng)
            else:
                result = replace_hints_without_precision_with_jokes(full_hints_patches)
        else:
            result = patches

        results[player_state.index] = FillerPlayerResult(
            game=game,
            patches=result,
            unassigned_pickups=player_state.pickups_left + player_expansions[player_state.index],
        )

    return FillerResults(results, actions_log)



import asyncio
import random
from argparse import ArgumentParser

from randovania.layout.permalink import Permalink


async def permalink_command_body(args):
    from randovania.interface_common import persistence
    from randovania.interface_common.preset_manager import PresetManager

    preset_manager = PresetManager(persistence.user_data_dir())
    await preset_manager.load_user_presets()

    versioned_preset = preset_manager.preset_for_name(args.preset)
    if versioned_preset is None:
        raise ValueError(f"Unknown preset: {args.preset}")

    seed = args.seed
    if seed is None:
        seed = random.randint(0, 2 ** 31)

    preset = versioned_preset.get_preset()
    permalink = Permalink(
        seed_number=seed,
        spoiler=not args.race,
        presets={i: preset for i in range(args.player_count)}
    )
    print(permalink.as_base64_str)

    Permalink.from_str(permalink.as_base64_str)


def permalink_command(args):
    asyncio.run(permalink_command_body(args))


def add_permalink_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "permalink",
        help="Creates a permalink"
    )
    parser.add_argument("--preset", type=str, required=True, help="The name of the preset to use.")
    parser.add_argument("--player-count", type=int, default=1, help="The number of players in the permalink.")
    parser.add_argument("--seed", type=int, help="The seed number.")
    parser.add_argument("--race", default=False, action="store_true", help="Make a race permalink (without spoiler).")

    parser.set_defaults(func=permalink_command)



from dataclasses import dataclass
from enum import Enum
from typing import Iterator, Tuple, FrozenSet

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder, BitPackEnum
from randovania.game_description import default_database
from randovania.game_description.area_location import AreaLocation
from randovania.games.game import RandovaniaGame


class StartingLocationConfiguration(BitPackEnum, Enum):
    SHIP = "ship"
    RANDOM_SAVE_STATION = "random-save-station"
    CUSTOM = "custom"


def _areas_list(game: RandovaniaGame):
    world_list = default_database.game_description_for(game).world_list
    areas = [
        AreaLocation(world.world_asset_id, area.area_asset_id)
        for world in world_list.worlds
        for area in world.areas
        if area.valid_starting_location
    ]
    return list(sorted(areas))


@dataclass(frozen=True)
class StartingLocation(BitPackValue):
    locations: FrozenSet[AreaLocation]
    game: RandovaniaGame

    @classmethod
    def with_elements(cls, elements: Iterator[AreaLocation], game: RandovaniaGame) -> "StartingLocation":
        return cls(frozenset(sorted(elements)), game)

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        yield from bitpacking.pack_sorted_array_elements(list(sorted(self.locations)), _areas_list(self.game))

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "StartingLocation":
        game = metadata["reference"].game
        return cls.with_elements(bitpacking.decode_sorted_array_elements(decoder, _areas_list(game)), game)

    @property
    def as_json(self) -> list:
        world_list = default_database.game_description_for(self.game).world_list
        return list(sorted(
            world_list.area_name(world_list.area_by_area_location(location), separator="/",
                                 distinguish_dark_aether=False)
            for location in self.locations
        ))

    @classmethod
    def from_json(cls, value: list, game: RandovaniaGame) -> "StartingLocation":
        if not isinstance(value, list):
            raise ValueError("StartingLocation from_json must receive a list, got {}".format(type(value)))

        world_list = default_database.game_description_for(game).world_list

        elements = []
        for location in value:
            world_name, area_name = location.split("/")
            world = world_list.world_with_name(world_name)
            area = world.area_by_name(area_name)
            if area.valid_starting_location:
                elements.append(AreaLocation(world.world_asset_id, area.area_asset_id))

        return cls.with_elements(elements, game)

    def ensure_has_location(self, area_location: AreaLocation, enabled: bool) -> "StartingLocation":
        new_locations = set(self.locations)
        if enabled:
            new_locations.add(area_location)
        elif area_location in new_locations:
            new_locations.remove(area_location)

        return StartingLocation.with_elements(iter(new_locations), self.game)

    def ensure_has_locations(self, area_locations: list, enabled: bool) -> "StartingLocation":
        new_locations = set(self.locations)
        for area_location in area_locations:
            if enabled:
                new_locations.add(area_location)
            elif area_location in new_locations:
                new_locations.remove(area_location)
        return StartingLocation.with_elements(iter(new_locations), self.game)




import sys
from multiprocessing import connection
from unittest.mock import patch, MagicMock, ANY, call

import pytest

from randovania.games.prime import iso_packager

work_around_pytest_qt_bug = False  # TODO: pytest-qt bug


@patch("nod.ExtractionContext", autospec=work_around_pytest_qt_bug)
@patch("nod.open_disc_from_image", autospec=work_around_pytest_qt_bug)
def test_disc_unpack_process_valid(mock_open_disc_from_image: MagicMock,
                                   mock_extraction_context: MagicMock,
                                   ):
    # TODO: invalid ISO

    # Setup
    output_pipe = MagicMock()
    iso = MagicMock()
    game_files_path = MagicMock()

    disc = MagicMock()
    mock_open_disc_from_image.return_value = disc, False
    data_partition = disc.get_data_partition.return_value

    def set_progress_side_effect(callable):
        callable("first path", 0.5)

    context = mock_extraction_context.return_value
    context.set_progress_callback.side_effect = set_progress_side_effect

    # Run
    iso_packager._disc_unpack_process(output_pipe, iso, game_files_path)

    # Assert
    mock_open_disc_from_image.assert_called_once_with(str(iso))
    disc.get_data_partition.assert_called_once_with()
    mock_extraction_context.assert_called_once_with()
    context.set_progress_callback.assert_called_once_with(ANY)
    data_partition.extract_to_directory.assert_called_once_with(str(game_files_path), context)

    output_pipe.send.assert_has_calls([
        call((False, "first path", 0.5)),
        call((True, None, 100)),
    ])


@patch("nod.ExtractionContext", autospec=work_around_pytest_qt_bug)
@patch("nod.open_disc_from_image", autospec=work_around_pytest_qt_bug)
def test_disc_unpack_process_invalid(mock_open_disc_from_image: MagicMock,
                                     mock_extraction_context: MagicMock,
                                     ):
    # Setup
    output_pipe = MagicMock()
    iso = MagicMock()
    game_files_path = MagicMock()

    disc = MagicMock()
    mock_open_disc_from_image.return_value = disc, False
    disc.get_data_partition.return_value = None

    # Run
    iso_packager._disc_unpack_process(output_pipe, iso, game_files_path)

    # Assert
    mock_open_disc_from_image.assert_called_once_with(str(iso))
    disc.get_data_partition.assert_called_once_with()
    mock_extraction_context.assert_not_called()

    output_pipe.send.assert_called_once_with(
        (True, "Could not find a data partition in '{}'.\nIs it a valid Metroid Prime 2 ISO?".format(iso), 0)
    )


@patch("nod.DiscBuilderGCN", autospec=work_around_pytest_qt_bug)
def test_disc_pack_process_success(mock_disc_builder: MagicMock,
                                   ):
    # Setup
    status_queue = MagicMock()
    iso = MagicMock()
    game_files_path = MagicMock()
    disc_builder = mock_disc_builder.return_value

    def builder_side_effect(_, callback):
        callback(0.5, "file one", 123456)
        return disc_builder

    mock_disc_builder.side_effect = builder_side_effect

    # Run
    iso_packager._disc_pack_process(status_queue, iso, game_files_path)

    # Assert
    mock_disc_builder.assert_called_once_with(str(iso), ANY)
    disc_builder.build_from_directory.assert_called_once_with(str(game_files_path))
    status_queue.send.assert_has_calls([
        call((False, "file one", 0.5)),
        call((True, None, 100)),
    ])


@patch("nod.DiscBuilderGCN", autospec=work_around_pytest_qt_bug)
def test_disc_pack_process_failure(mock_disc_builder: MagicMock,
                                   ):
    # Setup
    status_queue = MagicMock()
    iso = MagicMock()
    game_files_path = MagicMock()

    mock_disc_builder.side_effect = RuntimeError("Failure to do stuff")

    # Run
    iso_packager._disc_pack_process(status_queue, iso, game_files_path)

    # Assert
    mock_disc_builder.assert_called_once_with(str(iso), ANY)
    status_queue.send.assert_called_once_with((True, "Failure to do stuff", 0))


def test_shared_process_code_success():
    # Setup
    mock_target = MagicMock()
    iso = MagicMock()
    game_files_path = MagicMock()
    on_finish_message = MagicMock()
    progress_update = MagicMock()

    process = MagicMock()

    def process_effect(target, args):
        output_pipe: connection.Connection = args[0]
        output_pipe.send((False, "Message 1", 0.2))
        output_pipe.send((False, "Message 2", 0.5))
        output_pipe.send((True, None, 100))
        return process

    # Run
    with patch("multiprocessing.Process", side_effect=process_effect,
               autospec=work_around_pytest_qt_bug) as mock_process:
        iso_packager._shared_process_code(mock_target, iso, game_files_path, on_finish_message, progress_update)

    mock_process.assert_called_once_with(target=mock_target, args=(ANY, iso, game_files_path))
    process.start.assert_called_once_with()
    process.terminate.assert_called_once_with()
    progress_update.assert_has_calls([
        call("", 0),
        call("Message 1", 0.2),
        call("Message 2", 0.5),
        call(on_finish_message, 1),
    ])


def test_shared_process_code_failure():
    # Setup
    mock_target = MagicMock()
    iso = MagicMock()
    game_files_path = MagicMock()
    on_finish_message = MagicMock()
    progress_update = MagicMock()

    process = MagicMock()

    def process_effect(target, args):
        output_pipe: connection.Connection = args[0]
        output_pipe.send((False, "Message 1", 0.2))
        output_pipe.send((True, "You got an error!", 100))
        return process

    # Run
    with patch("multiprocessing.Process", side_effect=process_effect,
               autospec=work_around_pytest_qt_bug) as mock_process:
        with pytest.raises(RuntimeError) as exception:
            iso_packager._shared_process_code(mock_target, iso, game_files_path, on_finish_message, progress_update)

    mock_process.assert_called_once_with(target=mock_target, args=(ANY, iso, game_files_path))
    process.start.assert_called_once_with()
    process.terminate.assert_called_once_with()
    progress_update.assert_has_calls([
        call("", 0),
        call("Message 1", 0.2),
    ])
    assert str(exception.value) == "You got an error!"


@patch("randovania.games.prime.iso_packager._shared_process_code", autospec=True)
def test_unpack_iso_success(mock_shared_process_code: MagicMock,
                            ):
    # Setup
    iso = MagicMock()
    game_files_path = MagicMock()
    progress_update = MagicMock()

    # Run
    iso_packager.unpack_iso(iso, game_files_path, progress_update)

    # Assert
    game_files_path.mkdir.assert_called_once_with(parents=True, exist_ok=True)
    mock_shared_process_code.assert_called_once_with(
        target=iso_packager._disc_unpack_process,
        iso=iso,
        game_files_path=game_files_path,
        on_finish_message="Finished extracting ISO",
        progress_update=progress_update
    )


@patch("randovania.games.prime.iso_packager._shared_process_code", autospec=True)
def test_unpack_iso_failure(mock_shared_process_code: MagicMock,
                            ):
    # Setup
    iso = MagicMock()
    game_files_path = MagicMock()
    progress_update = MagicMock()
    exception_message = "Nah, don't wanna"

    game_files_path.mkdir.side_effect = OSError(exception_message)

    # Run
    with pytest.raises(RuntimeError) as exception:
        iso_packager.unpack_iso(iso, game_files_path, progress_update)

    # Assert
    game_files_path.mkdir.assert_called_once_with(parents=True, exist_ok=True)
    mock_shared_process_code.assert_not_called()
    assert str(exception.value) == "Unable to create files dir {}:\n{}".format(game_files_path, exception_message)


@pytest.mark.parametrize("iso_too_big", [False, True])
@patch("randovania.games.prime.iso_packager.nod")
@patch("randovania.games.prime.iso_packager._shared_process_code", autospec=True)
@patch("randovania.games.prime.iso_packager.validate_game_files_path", autospec=True)
def test_pack_iso(mock_validate_game_files_path: MagicMock,
                  mock_shared_process_code: MagicMock,
                  mock_nod: MagicMock,
                  iso_too_big: bool):
    # Setup
    iso = MagicMock()
    game_files_path = MagicMock()
    progress_update = MagicMock()

    sizes = [None if iso_too_big else 1]

    mock_calculate_total_size_required = mock_nod.DiscBuilderGCN.calculate_total_size_required
    mock_calculate_total_size_required.side_effect = sizes
    mock_nod.VERSION = "1.1.0"

    def run():
        iso_packager.pack_iso(iso, game_files_path, progress_update)

    # Run
    if iso_too_big:
        with pytest.raises(RuntimeError) as exception:
            run()
        assert str(exception.value) == "Image built with given directory would pass the maximum size."
    else:
        run()

    # Assert
    mock_validate_game_files_path.assert_called_once_with(game_files_path.joinpath.return_value)

    mock_calculate_total_size_required.assert_has_calls([
        call(str(game_files_path))
        for _ in sizes
    ])

    if iso_too_big:
        mock_shared_process_code.assert_not_called()
    else:
        mock_shared_process_code.assert_called_once_with(
            target=iso_packager._disc_pack_process,
            iso=iso,
            game_files_path=game_files_path,
            on_finish_message="Finished packing ISO",
            progress_update=progress_update
        )


@patch("randovania.games.prime.iso_packager.nod")
@patch("randovania.games.prime.iso_packager.validate_game_files_path", autospec=True)
def test_pack_iso_invalid_version(mock_validate_game_files_path: MagicMock,
                                  mock_nod: MagicMock,
                                  ):
    # Setup
    game_files_path = MagicMock()
    mock_nod.VERSION = "1.0.0"

    # Run
    with pytest.raises(RuntimeError) as exception:
        iso_packager.pack_iso(None, game_files_path, None)

    # Assert
    mock_validate_game_files_path.assert_called_once_with(game_files_path.joinpath.return_value)
    assert str(exception.value) == "Installed nod version (1.0) is older than required 1.1.0"


def test_can_process_iso_success():
    assert iso_packager.can_process_iso()


def test_can_process_iso_failure():
    with patch.dict(sys.modules, nod=None):
        del sys.modules["randovania.games.prime.iso_packager"]
        from randovania.games.prime.iso_packager import can_process_iso
        assert not can_process_iso()



import pytest
from mock import MagicMock, AsyncMock

from randovania.gui.dialog.login_prompt_dialog import LoginPromptDialog


@pytest.fixture(name="prompt")
def login_prompt_fixture(skip_qtbot):
    return LoginPromptDialog(MagicMock())


@pytest.mark.asyncio
async def test_on_login_as_guest_button(prompt, mocker):
    # Setup
    prompt.network_client.login_as_guest = AsyncMock()
    mock_dialog = mocker.patch("PySide2.QtWidgets.QInputDialog").return_value
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=mock_dialog.Accepted)

    # Run
    await prompt.on_login_as_guest_button()

    # Assert
    mock_execute_dialog.assert_awaited_once_with(mock_dialog)
    prompt.network_client.login_as_guest.assert_awaited_once_with(mock_dialog.textValue.return_value)


@pytest.mark.asyncio
async def test_on_login_with_discord_button(prompt):
    # Setup
    prompt.network_client.login_with_discord = AsyncMock()

    # Run
    await prompt.on_login_with_discord_button()

    # Assert
    prompt.network_client.login_with_discord.assert_awaited_once_with()



from typing import NamedTuple, Tuple

from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_type import ResourceType


class DamageReduction(NamedTuple):
    inventory_item: ItemResourceInfo
    damage_multiplier: float


class DamageResourceInfo(NamedTuple):
    index: int
    long_name: str
    short_name: str
    reductions: Tuple[DamageReduction, ...]

    @property
    def resource_type(self) -> ResourceType:
        return ResourceType.DAMAGE

    def damage_reduction(self, current_resources: "CurrentResources") -> float:
        multiplier = 1

        for reduction in self.reductions:
            if current_resources.get(reduction.inventory_item, 0) > 0:
                multiplier *= reduction.damage_multiplier

        return multiplier

    def __str__(self):
        return "{} Damage".format(self.long_name)



import logging
import struct
from typing import List, Optional, Dict

from PySide2 import QtWidgets
from PySide2.QtWidgets import QMainWindow
from asyncqt import asyncSlot

from randovania.game_connection.connection_backend import ConnectionBackend, MemoryOperation, _powerup_offset
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_description.node import PickupNode
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.games.prime import dol_patcher
from randovania.gui.generated.debug_backend_window_ui import Ui_DebugBackendWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.qt_network_client import handle_network_errors
from randovania.interface_common import enum_lib
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.network_common.admin_actions import SessionAdminUserAction


class DebugGameBackendChoice:
    @property
    def pretty_text(self):
        return "Debug"


class DebugBackendWindow(ConnectionBackend, Ui_DebugBackendWindow):
    pickups: List[PickupEntry]
    permanent_pickups: List[PickupEntry]

    def __init__(self):
        super().__init__()
        self.logger.setLevel(logging.DEBUG)
        self.window = QMainWindow()
        self.setupUi(self.window)
        common_qt_lib.set_default_window_icon(self.window)

        for status in enum_lib.iterate_enum(GameConnectionStatus):
            self.current_status_combo.addItem(status.pretty_text, status)

        self.permanent_pickups = []
        self.pickups = []

        self.collect_location_combo.setVisible(False)
        self.setup_collect_location_combo_button = QtWidgets.QPushButton(self.window)
        self.setup_collect_location_combo_button.setText("Load list of locations")
        self.setup_collect_location_combo_button.clicked.connect(self._setup_locations_combo)
        self.gridLayout.addWidget(self.setup_collect_location_combo_button, 1, 0, 1, 1)

        self.collect_location_button.clicked.connect(self._emit_collection)
        self.collect_location_button.setEnabled(False)

        self._expected_patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
        # FIXME: use PAL again
        self.patches = self._expected_patches

        self._game_memory = bytearray(24 * (2 ** 20))
        self._game_memory_initialized = False
        self.patches = None

    async def _ensure_initialized_game_memory(self):
        if self._game_memory_initialized:
            return
        try:
            self.patches = self._expected_patches
            world = self.game.world_list.worlds[0]

            await self._perform_memory_operations([
                # Build String
                MemoryOperation(self.patches.build_string_address, write_bytes=self._expected_patches.build_string),

                # current CWorld
                MemoryOperation(self.patches.game_state_pointer, offset=4,
                                write_bytes=world.world_asset_id.to_bytes(4, "big")),

                # CPlayer VTable
                MemoryOperation(self.patches.cstate_manager_global + 0x14fc, offset=0,
                                write_bytes=self.patches.cplayer_vtable.to_bytes(4, "big")),

                # CPlayerState
                MemoryOperation(self.patches.cstate_manager_global + 0x150c,
                                write_bytes=0xA00000.to_bytes(4, "big")),
            ])

            self._game_memory_initialized = True
        finally:
            self.patches = None

    def _read_memory(self, address: int, count: int):
        address &= ~0x80000000
        return self._game_memory[address:address + count]

    def _read_memory_format(self, format_string: str, address: int):
        return struct.unpack_from(format_string, self._game_memory, address & ~0x80000000)

    def _write_memory(self, address: int, data: bytes):
        address &= ~0x80000000
        self._game_memory[address:address + len(data)] = data
        self.logger.info(f"Wrote {data.hex()} to {hex(address)}")

    @property
    def current_status(self) -> GameConnectionStatus:
        return self.current_status_combo.currentData()

    @property
    def backend_choice(self):
        return DebugGameBackendChoice()

    @property
    def lock_identifier(self) -> Optional[str]:
        return None

    @property
    def name(self) -> str:
        return "Debug"

    async def update(self, dt: float):
        await self._ensure_initialized_game_memory()

        if not self._enabled:
            return

        if not await self._identify_game():
            return

        await self._interact_with_game(dt)
        self._read_message_from_game()
        self._update_inventory_label()

    def _update_inventory_label(self):
        s = "<br />".join(
            f"{name} x {quantity.amount}/{quantity.capacity}" for name, quantity in self._inventory.items()
        )
        self.inventory_label.setText(s)

    def show(self):
        self.window.show()

    def _get_magic_address(self):
        multiworld_magic_item = self.game.resource_database.multiworld_magic_item
        magic_address = 0xA00000 + _powerup_offset(multiworld_magic_item.index)
        return magic_address

    def _read_magic(self):
        return self._read_memory_format(">II", self._get_magic_address())

    def _write_magic(self, magic_amount, magic_capacity):
        self._write_memory(self._get_magic_address(), struct.pack(">II", magic_amount, magic_capacity))

    @asyncSlot()
    async def _emit_collection(self):
        new_magic_value = self.collect_location_combo.currentData() + 1
        magic_amount, magic_capacity = self._read_magic()
        magic_amount += new_magic_value
        magic_capacity += new_magic_value
        self._write_magic(magic_amount, magic_capacity)

    @asyncSlot()
    @handle_network_errors
    async def _setup_locations_combo(self):
        network_client = common_qt_lib.get_network_client()
        game_session = network_client.current_game_session
        user = network_client.current_user

        game = self.game
        index_to_name = {
            node.pickup_index.index: game.world_list.area_name(area)
            for world, area, node in game.world_list.all_worlds_areas_nodes
            if isinstance(node, PickupNode)
        }

        if game_session is None:
            names = index_to_name
        else:
            patcher_data = await network_client.session_admin_player(user.id,
                                                                     SessionAdminUserAction.CREATE_PATCHER_FILE,
                                                                     CosmeticPatches().as_json)
            names = {
                pickup["pickup_index"]: "{}: {}".format(index_to_name[pickup["pickup_index"]],
                                                        pickup["hud_text"][0])
                for pickup in patcher_data["pickups"]
            }

        self.collect_location_combo.clear()
        for index, name in sorted(names.items()):
            self.collect_location_combo.addItem(name, index)

        self.collect_location_button.setEnabled(True)
        self.collect_location_combo.setVisible(True)
        self.setup_collect_location_combo_button.deleteLater()

    def clear(self):
        self.messages_list.clear()
        self.permanent_pickups = []
        self.pickups.clear()

    def _memory_operation(self, op: MemoryOperation) -> Optional[bytes]:
        op.validate_byte_sizes()

        address = op.address
        if op.offset is not None:
            address = self._read_memory_format(">I", address)[0]
            address += op.offset

        result = None
        if op.read_byte_count is not None:
            result = self._read_memory(address, op.read_byte_count)
        if op.write_bytes is not None:
            self._write_memory(address, op.write_bytes)
        return result

    async def _perform_memory_operations(self, ops: List[MemoryOperation]) -> Dict[MemoryOperation, bytes]:
        result = {}
        for op in ops:
            op_result = self._memory_operation(op)
            if op_result is not None:
                result[op] = op_result
        return result

    def _read_message_from_game(self):
        has_message_address = self.patches.cstate_manager_global + 0x2
        if self._read_memory(has_message_address, 1) == b"\x00":
            return

        string_start = self.patches.string_display.message_receiver_string_ref
        message_bytes = self._read_memory(string_start, self.patches.string_display.max_message_size + 2)
        message = message_bytes[:message_bytes.find(b"\x00\x00")].decode("utf-16_be")

        self.messages_list.addItem(message)
        self._write_memory(has_message_address, b"\x00")
        self._write_magic(0, len(self.permanent_pickups))



import json
import pickle
from pathlib import Path

import pytest

from randovania.layout.layout_description import LayoutDescription
from randovania.layout.trick_level import LayoutTrickLevel


@pytest.mark.parametrize("value", LayoutTrickLevel)
def test_pickle_trick_level(value: LayoutTrickLevel):
    assert pickle.loads(pickle.dumps(value)) == value


def test_load_multiworld(test_files_dir):
    file_path = Path(test_files_dir).joinpath("log_files", "multiworld.rdvgame")

    with file_path.open("r") as open_file:
        input_data = json.load(open_file)

    # Run
    result = LayoutDescription.from_json_dict(input_data)

    # Assert
    as_json = result.as_json
    del input_data["info"]
    del as_json["info"]

    assert as_json == input_data



from typing import Union, Tuple, Iterator, Dict

from randovania.game_description.resources.damage_resource_info import DamageResourceInfo
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo

ResourceInfo = Union[SimpleResourceInfo, ItemResourceInfo, TrickResourceInfo,
                     DamageResourceInfo, PickupIndex, TranslatorGate, LogbookAsset]
ResourceQuantity = Tuple[ResourceInfo, int]
ResourceGainTuple = Tuple[ResourceQuantity, ...]
ResourceGain = Iterator[ResourceQuantity]
CurrentResources = Dict[ResourceInfo, int]


def add_resource_gain_to_current_resources(resource_gain: ResourceGain,
                                           resources: CurrentResources) -> CurrentResources:
    """
    Adds all resources from the given gain to the given CurrentResources
    :param resource_gain:
    :param resources:
    :return: resources
    """
    for resource, quantity in resource_gain:
        resources[resource] = resources.get(resource, 0) + quantity
    return resources


def add_resources_into_another(target: CurrentResources, source: CurrentResources) -> None:
    resource_gain: ResourceGain = source.items()
    add_resource_gain_to_current_resources(resource_gain, target)


def convert_resource_gain_to_current_resources(resource_gain: ResourceGain) -> CurrentResources:
    """
    Creates a CurrentResources with all resources of the given ResourceGain
    :param resource_gain:
    :return:
    """
    result = {}
    add_resource_gain_to_current_resources(resource_gain, result)
    return result



import multiprocessing

from PySide2 import QtCore, QtWidgets
from asyncqt import asyncSlot

from randovania.generator.filler.filler_library import UnableToGenerate
from randovania.gui.lib import async_dialog, common_qt_lib
from randovania.resolver.exceptions import GenerationFailure, ImpossibleForSolver


class GenerationFailureHandler(QtWidgets.QWidget):
    failed_to_generate_signal = QtCore.Signal(GenerationFailure)

    def __init__(self, parent: QtWidgets.QWidget):
        super().__init__(parent)

        self.parent = parent
        self.failed_to_generate_signal.connect(self._show_failed_generation_exception)

    def handle_failure(self, failure: GenerationFailure):
        self.failed_to_generate_signal.emit(failure)

    @asyncSlot(GenerationFailure)
    async def _show_failed_generation_exception(self, exception: GenerationFailure):
        box = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Critical,
            "An error occurred while generating game for {0.permalink.as_base64_str}".format(exception),
            str(exception), QtWidgets.QMessageBox.Ok, self.parent)
        common_qt_lib.set_default_window_icon(box)

        if isinstance(exception.source, UnableToGenerate):
            box.setText(box.text() + "\n\nClick 'Show Details' to see a report of where the failure occurred.\n"
                                     "Double check if your settings aren't impossible, or try again.")
            box.setDetailedText(str(exception.source))

        elif isinstance(exception.source, ImpossibleForSolver):
            box.setText(box.text() + "\n\nRandovania sometimes generates games with insufficient Energy Tanks. "
                                     "Please try generating again.")

        elif isinstance(exception.source, multiprocessing.TimeoutError):
            box.setText(box.text() + "\n\nRandovania sometimes gets stuck infinitely when trying to verify a game, "
                                     "so there's a timeout. Please try generating again.")

        await async_dialog.execute_dialog(box)



import functools
import json
from pathlib import Path
from typing import Tuple

from randovania import get_data_path
from randovania.games.game import RandovaniaGame
from randovania.games.prime.binary_data import decode_file_path


@functools.lru_cache()
def read_json_then_binary(game: RandovaniaGame) -> Tuple[Path, dict]:
    json_path = get_data_path().joinpath("json_data", f"{game.value}.json")
    if json_path.exists():
        with json_path.open("r") as open_file:
            return json_path, json.load(open_file)

    binary_path = get_data_path().joinpath("binary_data", f"{game.value}.bin")
    return binary_path, decode_file_path(binary_path)


def decode_default_prime2() -> dict:
    return read_json_then_binary(RandovaniaGame.PRIME2)[1]


def decode_default_prime3() -> dict:
    return read_json_then_binary(RandovaniaGame.PRIME3)[1]


@functools.lru_cache()
def decode_randomizer_data() -> dict:
    randomizer_data_path = get_data_path().joinpath("ClarisPrimeRandomizer", "RandomizerData.json")

    with randomizer_data_path.open() as randomizer_data_file:
        return json.load(randomizer_data_file)



import pytest
from PySide2.QtCore import Qt

from randovania.games.game import RandovaniaGame
from randovania.gui.preset_settings.logic_settings_window import LogicSettingsWindow
from randovania.interface_common.preset_editor import PresetEditor


@pytest.mark.parametrize("game", [RandovaniaGame.PRIME2, RandovaniaGame.PRIME3])
def test_on_preset_changed(skip_qtbot, preset_manager, game):
    # Setup
    editor = PresetEditor(preset_manager.default_preset_for_game(game).get_preset())
    window = LogicSettingsWindow(None, editor)

    # Run
    window.on_preset_changed(editor.create_custom_preset_with())


def test_starting_location_world_select(skip_qtbot, default_preset):
    # Setup
    editor = PresetEditor(default_preset)
    window = LogicSettingsWindow(None, editor)
    skip_qtbot.addWidget(window)

    # Run
    checkbox_list = window._starting_location_for_world
    window.on_preset_changed(editor.create_custom_preset_with())
    assert len(checkbox_list) == 10
    temple_grounds_checkbox = checkbox_list["Temple Grounds"]
    assert temple_grounds_checkbox.checkState() == Qt.PartiallyChecked
    skip_qtbot.mouseClick(temple_grounds_checkbox, Qt.LeftButton)
    assert temple_grounds_checkbox.checkState() == Qt.Checked
    assert len(editor.configuration.starting_location.locations) == 39
    skip_qtbot.mouseClick(temple_grounds_checkbox, Qt.LeftButton)
    assert temple_grounds_checkbox.checkState() == Qt.Unchecked
    assert len(editor.configuration.starting_location.locations) == 0
    skip_qtbot.mouseClick(temple_grounds_checkbox, Qt.LeftButton)
    window.on_preset_changed(editor.create_custom_preset_with())
    assert temple_grounds_checkbox.checkState() == Qt.Checked
    assert len(editor.configuration.starting_location.locations) == 39
